#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import sys
import re
import traceback
import glob
import numpy as np

from casacore.tables.tableutil import tableexists
import Owlcat.Tables

flagger = parser = ms = msname = None


def error(message):
    sys.stderr.write("%s: %s\n" % (os.path.basename(sys.argv[0]), message))
    sys.exit(1)


def get_ms(readonly=True):
    global ms
    global msname
    if not ms:
        ms = Owlcat.table(msname, readonly=readonly, ack=False)
    return ms


def shape_str(label, arr):
    return "%s: %s" % (label, "x".join(map(str, arr.shape)) if arr is not None else "None")


def parse_subset_options(options):
    global parser
    global flagger
    global msname
    global ms
    subset = {}
    from Owlcat import Parsing

    # DDID and FIELD_ID
    if options.ddid is not None:
        try:
            subset['ddid'] = list(map(int, options.ddid.split(",")))
            print("  ===> DATA_DESC_ID:", subset['ddid'])
        except:
            parser.error("Invalid -D/--ddid option")
    if options.field is not None:
        try:
            subset['fieldid'] = list(map(int, options.field.split(",")))
            print("  ===> FIELD_ID:", subset['fieldid'])
        except:
            parser.error("Invalid -F/--field option")
    # taql
    taqls = []
    if options.taql:
        taqls.append(options.taql)
        print("  ===> TaQL query:", options.taql)
    # channels
    if options.channels:
        subset['channels'] = list(map(Parsing.parse_slice, options.channels.split(",")))
        print("  ===> channels:", subset['channels'])
    # corr list
    if options.corrs is not None:
        try:
            subset['corrs'] = list(map(int, options.corrs.split(',')))
            print("  ===> correlations:", subset['corrs'])
        except:
            parser.error("Invalid -X/--corrs option")
    # station list
    if options.stations is not None:
        try:
            subset['antennas'] = list(map(int, options.stations.split(',')))
            print("  ===> stations:", subset['antennas'])
        except:
            parser.error("Invalid -S/--stations option")
    # IFR set
    if options.ifrs is not None:
        from Cattery.Meow import IfrSet
        ifrset = IfrSet.from_ms(get_ms())
        # print help and exit
        if options.ifrs == "help":
            # print help string, but trim away RTF tags
            print(re.sub("<[^>]+>", "", ifrset.subset_doc).replace("&lt;", "<").replace("&gt;", ">"))
            sys.exit(0)
        try:
            ifrset = ifrset.subset(options.ifrs)
            print("  ===> ifrs:", " ".join([ifrset.ifr_label(ip, iq) for (ip, p), (iq, q) in ifrset.ifr_index()]))
            if not ifrset.ifrs():
                return None
        except:
            parser.error("Invalid -I/--ifrs option")
        taqls.append(ifrset.taql_string())
    # clipping
    subset['data_column'] = options.data_column
    if options.nan:
        subset['data_nan'] = True
        print("  ===> select %s = NAN or INF" % options.data_column)
    if options.above is not None:
        subset['data_above'] = options.above
        print("  ===> select |%s|>%f" % (options.data_column, options.above))
    if options.below is not None:
        subset['data_below'] = options.below
        print("  ===> select |%s|<%f" % (options.data_column, options.below))
    if options.fm_above is not None:
        subset['data_fm_above'] = options.fm_above
        print("  ===> select mean|%s|>%f" % (options.data_column, options.fm_above))
    if options.fm_below is not None:
        subset['data_fm_below'] = options.fm_below
        print("  ===> select mean|%s|<%f" % (options.data_column, options.fm_below))
    # join taql queries
    if taqls:
        subset['taql'] = "( " + " ) && ( ".join(taqls) + " )"
    # fill flag args
    for opt in 'data_flagmask', 'flagmask', 'flagmask_all', 'flagmask_none':
        subset[opt] = getattr(options, opt)
    return subset

def copy_chunked_flags(tab_from, tab_to, chunk_size):
    """copies FLAG column in a chunked way. Returns two sums"""
    nrows = tab_from.nrows()
    fc_sum = fc_size = 0
    for row0 in range(0, nrows, chunk_size):
        n = min(nrows - row0, chunk_size)
        fr = tab_from.getcol("FLAG_ROW", row0, n)
        fc = tab_from.getcol("FLAG", row0, n)
        fc[fr,:,:] = True
        if tab_to is not None:
            tab_to.putcol("FLAG_ROW", fr, row0, n)
            tab_to.putcol("FLAG", fc, row0, n)
        fc_sum += fc.sum(dtype=np.int64)
        fc_size += fc.size
    return fc_sum / float(fc_size)


if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup

    parser = OptionParser(usage="""%prog: [actions] [options] MS""",
                          description="Manipulates flags (bitflags and legacy FLAG/FLAG_ROW columns) in the MS. "
                                      "Use the selection options to narrow down a subset of the data, and use the action options "
                                      "to change flags within that subset. Without any action options, statistics on the current "
                                      "selection are printed -- this is useful as a preview of your intended action."
                          )

    group = OptionGroup(parser, "Selection by subset")
    group.add_option("-L", "--channels", type="string",
                     help="channel selection: single number or start:end[:step] to select channels start through end-1, "
                          "or start~end[:step] to select channels start through end, with an optional stepping.")
    group.add_option("-T", "--timeslots", type="string",
                     help="timeslot selection: single number or start:end to select timeslots start through end-1, "
                          "or start~end to select timeslots start through end.")
    group.add_option("-M", "--timeslot-multiplier", type="int", default=1,
                     help="multiplies the timeslot numbers given to -T by the given factor. Default is 1.")
    group.add_option("-X", "--corrs", type="string",
                     help="correlation selection. Use comma-separated list of correlation indices.")
    group.add_option("-S", "--stations", type="string",
                     help="station (=antenna) selection. Use comma-separated list of station indices."),
    group.add_option("-I", "--ifrs", type="string",
                     help="interferometer selection. Use \"-I help\" to get help on selecting ifrs.")
    group.add_option("-D", "--ddid", type="string",
                     help="DATA_DESC_ID selection. Single number, or comma-separated list.")
    group.add_option("-F", "--field", type="string",
                     help="FIELD_ID selection. Single number, or comma-separated list.")
    group.add_option("-Q", "--taql", dest="taql", type="str",
                     help="additional TaQL selection to restrict subset.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Selection by data value")
    group.add_option("--above", metavar="X", type="float",
                     help="select on abs(data)>X")
    group.add_option("--below", metavar="X", type="float",
                     help="select on abs(data)<X")
    group.add_option("--nan", action="store_true",
                     help="select on invalid data (NaN or infinite)")
    group.add_option("--fm-above", metavar="X", type="float",
                     help="select on mean(abs(data))>X, where mean is over frequencies")
    group.add_option("--fm-below", metavar="X", type="float",
                     help="select on mean(abs(data))<X, where mean is over frequencies")
    group.add_option("-C", "--data-column", metavar="COLUMN", type="string",
                     help="data column for --above/--below/--nan options. Default is %default.")
    group.add_option("--data-flagmask", metavar="FLAGS", type="string",
                     help="flags to apply to data column (when e.g. computing mean). Default is %default. See below for "
                          "details on specifying flags.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Selection by current flags")
    group.add_option("-Y", "--flagged-any", metavar="FLAGS", dest='flagmask', type="string",
                     help="selects if any of the specified flags are raised. For this and all other options taking "
                          "a FLAGS argument, FLAGS can be a flagset name or an integer bitmask "
                          "(if bitflags are in use -- see also the -l/--list option). Prefix the bitmask by '0x' to use hex. "
                          "Append a '+L' to include legacy boolean FLAG/FLAG_ROW columns. Use 'all' for "
                          "all bitflags, and 'ALL' for all bitflags plus legacy flags (equivalent to 'all+L'). FLAGS may "
                          "also be a comma-separated list of any of the above terms.")
    group.add_option("-A", "--flagged-all", metavar="FLAGS", dest='flagmask_all', type="string",
                     help="selects if all of the specified flags are raised")
    group.add_option("-N", "--flagged-none", metavar="FLAGS", dest='flagmask_none', type="string",
                     help="selects if none of the specified flags are raised")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Actions to take on selection (may be combined)")
    group.add_option("-x", "--extend-all-corr", action="store_true",
                     help="apply selection to all correlations if at least one is selected")
    group.add_option("-f", "--flag", metavar="FLAGS", type="string",
                     help="raise the specified FLAGS (selection added to output FLAGS)")
    group.add_option("-u", "--unflag", metavar="FLAGS", type="string",
                     help="clear the specified flags (selection removed from output FLAGS)")
    group.add_option("-y", "--copy", metavar="FLAGS", type="string",
                     help="copies to the the specified FLAGS (selection replaces output FLAGS)")
    group.add_option("--copy-legacy", metavar="FLAGS", type="string",
                     help="shortcut for --flagged-any +L --copy FLAGS --fill-legacy -")
    group.add_option("-g", "--fill-legacy", metavar="FLAGS", type="string",
                     help="fills legacy FLAG/FLAG_ROW columns using the specified FLAGS. When -f/--flag or -u/--unflag "
                          "or -r/--remove is used, legacy flags are implicitly reset using all bitflags: use '-g -' "
                          "to skip this step. You may also use this option on its own to reset legacy flags (within the "
                          "specified data subset) using some bitmask. Use '-g 0' to clear legacy flags.")
    group.add_option("-c", "--create", action="store_true",
                     help="for -f/--flag option only: if a named flagset doesn't exist, creates "
                          "it. Without this option, an error is reported.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Other options")
    group.add_option("--init-bitflags", type="choice", choices=["0", "8", "16", "32"], default=0,
                     help="initializes a BITFLAG column with the specified number of bits (8, 16 or 32) if it doesn't exist.")
    group.add_option("--reinit-bitflags", type="choice", choices=["0", "8", "16", "32"], default=0,
                     help="removes (if any) and reinitializes a BITFLAG column with the specified number of bits (8, 16 or 32) if it doesn't exist.")
    group.add_option("--incr-stman", action="store_true",
                     help="force the use of the incremental storage manager for new BITFLAG columns. Default is to use same manager as DATA column.")
    group.add_option("-l", "--list", action="store_true",
                     help="lists various info about the MS, including its flagsets, and CASA flagversions, if available.")
    group.add_option("-s", "--stats", action="store_true",
                     help="prints per-flagset flagging stats.")
    group.add_option("-r", "--remove", metavar="FLAGSET(s)", type="string",
                     help="unflags and removes named flagset(s). You can use a comma-separated list.")
    group.add_option("--export", type="string", metavar="FILENAME",
                     help="exports all flags to flag file. FILENAME may end with .gz to produce a gzip-compressed file. If any flagging actions are specified, these will be done before the export.")
    group.add_option("--import", type="string", dest="_import", metavar="FILENAME",
                     help="imports flags from flag file. If any flagging actions are specified, these will be done after the import.")
    group.add_option("-R", "--restore", type="string", dest="restore", metavar="VERSION",
                     help="restores flags from a CASA-style flagversion (use --list to list version names).")
    group.add_option("--save", type="string", dest="save", metavar="VERSION",
                     help="saves flags to a CASA-style flagversion.")
    group.add_option("--force", action="store_true",
                     help="with --save: force overwrite of a CASA-style flagversion, if it exists.")
    group.add_option("-v", "--verbose", metavar="LEVEL", type="int",
                     help="verbosity level for messages. Higher is more verbose, default is 0.")
    group.add_option("--timestamps", action="store_true",
                     help="adds timestamps to verbosity messages.")
    group.add_option("-z", "--chunk-size", metavar="NROWS", type="int", default=200000,
                     help="Number of rows to process at once. Default is %default. Set to higher values if you have RAM to spare.")
    parser.add_option_group(group)

    parser.set_defaults(data_column="CORRECTED_DATA", data_flagmask="ALL",
                        flagged_any=None, flaged_all=None, flagged_none=None,
                        flag=None, unflag=None, copy=None, copy_legacy=None, fill_legacy=None,
                        verbose=0)

    # parse args
    (options, args) = parser.parse_args()
    if len(args) != 1:
        parser.error("Incorrect number of arguments. Use '-h' for help.")
    msname = args[0].rstrip("/")

    import Owlcat

    # import flags from file, if so specified
    # these are the columns that are imported and exported
    FLAGCOLS = "FLAG", "FLAG_ROW", "BITFLAG", "BITFLAG_ROW"
    if options._import:
        try:
            dump = Owlcat.Tables.TableDump(options._import, compress=True)
            ms = get_ms(readonly=False)
            print("Importing flags from %s:" % options._import)
            dump.load(ms, verbose=True)
            dump.close()
            ms.close()
        except:
            traceback.print_exc()
            error("Error importing flags from %s, exiting" % options._import)
        print("Flags imported OK.")

    # same for CASA flagversions
    if options.restore:
        flagvers = f"{msname}.flagversions/flags.{options.restore}"
        if not Owlcat.tableexists(flagvers):
            error(f"No such flagversion '{options.restore}'. Use --list to list available flag versions.")
        try:
            ms = get_ms(readonly=False)
            ftab = Owlcat.table(flagvers)
            ratio = copy_chunked_flags(ftab, ms, options.chunk_size)
            ftab.close()
            ms.close()
            print(f"Flag version '{options.restore}' restored: {ratio:.2%} flagged.")
        except:
            traceback.print_exc()
            error(f"Error restoring flags from {options.restore}, exiting")

    # if no other actions supplied, enable stats (unless flags were imported, in which case just exit)
    if not (options.flag or options.unflag or options.copy or options.fill_legacy):
        if options._import or options.restore:
            sys.exit(0)
        statonly = True
    else:
        statonly = False

    import Owlcat.Flagger
    from Owlcat.Flagger import Flagger

    flagger = Flagger(msname, verbose=options.verbose, timestamps=options.timestamps, chunksize=options.chunk_size)

    # init/reinit bitflags
    init_bitflags = int(options.init_bitflags)
    reinit_bitflags = int(options.reinit_bitflags)
    init_nbits = reinit_bitflags or init_bitflags

    if init_nbits:
        if reinit_bitflags:
            print("{}: removing existing bitflag columns, if any".format(msname))
            flagger.remove_bitflags()
        else:
            if flagger.has_bitflags:
                if flagger.flagsets.NBITS != init_nbits:
                    error("this MS already has a bitflag column of an incompatible bitsize {}".format(flagger.flagsets.NBITS))
                else:
                    print("{}: bitflag columns seem to be in place".format(msname))
        if not flagger.has_bitflags:
            print("{}: inserting bitflag columns (size {} bits)".format(msname, init_nbits))
            stman = None
            if options.incr_stman:
                stman = "IncrementalStMan"
                print("  will use the {} storage manager".format(stman))
            else:
                print("  will use same storage manager as existing DATA column")
            print("{}: inserting bitflag columns (size {} bits)".format(msname, init_nbits))
            flagger.add_bitflags(bits=init_nbits, stman=stman)

    # now, skip most of the actions below if we're in statonly mode and exporting
    if not (statonly and options.export):
        #
        # -l/--list: list MS info
        #
        if options.list:
            ms = get_ms()
            ants = Owlcat.table(ms.getkeyword('ANTENNA'), ack=False).getcol('NAME')
            ddid_tab = Owlcat.table(ms.getkeyword('DATA_DESCRIPTION'), ack=False)
            spwids = ddid_tab.getcol('SPECTRAL_WINDOW_ID')
            polids = ddid_tab.getcol('POLARIZATION_ID')
            corrs = Owlcat.table(ms.getkeyword('POLARIZATION'), ack=False).getcol('CORR_TYPE')
            spw_tab = Owlcat.table(ms.getkeyword('SPECTRAL_WINDOW'), ack=False)
            ref_freq = spw_tab.getcol('REF_FREQUENCY')
            nchan = spw_tab.getcol('NUM_CHAN')
            fields = Owlcat.table(ms.getkeyword('FIELD'), ack=False).getcol('NAME')

            print("===> MS is %s" % msname)
            print("  %d antennas: %s" % (len(ants), " ".join(f"{num}:{name}" for num, name in enumerate(ants))))
            print("  %d DATA_DESC_ID(s): " % len(spwids))
            for i, (spw, pol) in enumerate(zip(spwids, polids)):
                print("    %d: %.3f MHz, %d chans x %d correlations" % (
                i, ref_freq[spw] * 1e-6, nchan[spw], len(corrs[pol, :])))
            print("  %d field(s): %s" % (len(fields), ", ".join(["%d: %s" % ff for ff in enumerate(fields)])))
            if not flagger.has_bitflags:
                print("No BITFLAG/BITFLAG_ROW columns in this MS. Use the 'addbitflagcol' command to add them.")
            else:
                names = flagger.flagsets.names()
                if names:
                    print("  %d flagset(s): " % len(names))
                    for name in names:
                        mask = flagger.flagsets.flagmask(name)
                        print("    '%s': %d (0x%02X)" % (name, mask, mask))
                else:
                    print("  No flagsets.")
            # now check for CASA flagversions
            flagvers = [tab for tab in glob.glob(f"{msname}.flagversions/flags.*") if Owlcat.tableexists(tab)]
            if flagvers:
                print(f"  {len(flagvers)} CASA-style flagversions available for --restore:")
                for vers in flagvers:
                    name = os.path.basename(vers).split('.', 1)[1]
                    print(f"    '{name}'")
            else:
                print("  No CASA-style flagversions available.")
            if options.flag or options.unflag or options.copy or options.fill_legacy or options.remove:
                print("-l/--list was in effect, so all other options were ignored.")
            print("")
            sys.exit(0)

        if options.copy_legacy:
            options.copy = options.copy_legacy
            options.fill_legacy = "-"
            options.flagmask = "+L"

        # --flag/--unflag/--remove implies '-g all' by default, '-g -' skips the fill-legacy step
        if options.flag or options.unflag or options.copy or options.remove:
            if options.fill_legacy is None:
                options.fill_legacy = 'all'
            elif options.fill_legacy == '-':
                options.fill_legacy = None

        # if no other actions supplied, enable stats (unless flags were imported, in which case just exit)
        if not (options.flag or options.unflag or options.copy or options.fill_legacy):
            if options._import:
                sys.exit(0)
            statonly = not options.export
        else:
            statonly = False

        # convert all the various FLAGS to flagmasks (or Nones)
        for opt in 'data_flagmask', 'flagmask', 'flagmask_all', 'flagmask_none', 'flag', 'copy', 'unflag', 'fill_legacy':
            value = getattr(options, opt)
            can_create = opt in ('flag', 'copy')
            # reopen for writing, if creating MS
            if can_create:
                flagger._reopen(True)
            try:
                flagmask = flagger.lookup_flagmask(value, create=can_create and options.create)
            except Exception as exc:
                msg = str(exc)
                if not options.create:
                    msg += "\nPerhaps you forgot the -c/--create option?"
                error(msg)
            setattr(options, opt, flagmask)

        ## # clear the legacy flag itself from fill_legacy, otherwise it can have no effect
        ## if options.fill_legacy is not None:
        ##    options.fill_legacy &= ~Flagger.LEGACY

        #
        # -r/--remove: remove flagsets
        #
        if options.remove is not None:
            if options.flag or options.unflag or options.copy:
                error("Can't combine -r/--remove with --f/--flag or -u/--unflag or -C/--copy")
            # get set of flagsets to remove
            remove_flagsets = set(options.remove.split(","))
            # get set of all flagsets
            all_flagsets = set(flagger.flagsets.names())
            # warn if any named flagsets were not found
            if remove_flagsets - all_flagsets:
                print("===> WARNING: flagset(s) %s not found, ignoring" % ",".join(remove_flagsets - all_flagsets))
            # build flagmask of remaining flagsets
            retain = all_flagsets - remove_flagsets
            if retain == all_flagsets:
                # if names_not_found:
                # error("No such flagset(s): %s"%",".join(names_not_found))
                print("===> WARNING: no flagsets to remove, exiting")
                sys.exit(0)
            flagmask = 0
            for name in retain:
                flagmask |= flagger.flagsets.flagmask(name)
            print("===> removing flagset(s) %s" % ",".join(all_flagsets - retain))
            print("===> and clearing flagmask %s" % flagger.flagmaskstr(~flagmask))
            if options.fill_legacy is not None:
                print("===> and filling FLAG/FLAG_ROW using flagmask %s" % flagger.flagmaskstr(options.fill_legacy))
            flagger.xflag(unflag=~flagmask, fill_legacy=options.fill_legacy)
            flagger.flagsets.remove_flagset(*list(all_flagsets - retain))
            sys.exit(0)

        # parse subset options
        subset = parse_subset_options(options)
        if not subset:
            print("===> ended up with empty subset, exiting")
            sys.exit(0)

        # convert timeslots to reltime option, if specified
        if options.timeslots:
            from Owlcat import Parsing

            tslice = Parsing.parse_slice(options.timeslots, options.timeslot_multiplier)
            times = sorted(set(get_ms().getcol('TIME')))
            time0 = times[0] if tslice.start is None else times[tslice.start]
            time1 = times[-1] if tslice.stop is None else times[tslice.stop - 1]
            time0 -= times[0]
            time1 -= times[0]
            subset['reltime'] = time0, time1
            print("  ===> select timeslots %s (reltime %g~%g s)" % (tslice, time0, time1))

        # at this stage all remaining options are handled the same way
        flagstr = unflagstr = legacystr = None
        if options.flag is not None:
            flagstr = flagger.flagmaskstr(options.flag)
            print("===> flagging with flagmask %s" % flagstr)
        if options.unflag is not None:
            unflagstr = flagger.flagmaskstr(options.unflag)
            print("===> unflagging with flagmask %s" % unflagstr)
        if options.copy is not None:
            copyflagstr = flagger.flagmaskstr(options.copy)
            print("===> copying to flagmask %s" % copyflagstr)
        if options.fill_legacy is not None:
            legacystr = flagger.flagmaskstr(options.fill_legacy)
            print("===> filling legacy flags with flagmask %s" % legacystr)

        # if --stats in effect, loop over all flagsets and print stats
        if options.stats:
            print("===> --stats in effect, showing per-flagset statistics")
            printed_header = False
            stats = {}
            if flagger.flagsets.names():
                for flagset in list(flagger.flagsets.names()):
                    stats[flagset] = flagger.lookup_flagmask(flagset)
            stats["+L"] = None

            totrows, sel_nrow, sel_nvis, nvis_A, nvis_B, nvis_C = flagger.xflag(get_stats_only=stats, **subset)
            percent = 100.0 / sel_nvis if sel_nvis else 0
            for flagset in list(flagger.flagsets.names() or []) + ["+L"]:
                nv = stats[flagset]
                # print them
                if flagset is "+L":
                    label = "legacy FLAG/FLAG_ROW"
                else:
                    label = "Flagset %s (0x%02X)" % (flagset, flagger.flagsets.flagmask(flagset))
                if not printed_header:
                    printed_header = True
                    rpc = 100.0 / totrows if totrows else 0
                    print("===>   MS size:               %8d rows" % totrows)
                    print("===>   Data/time selection:   %8d rows, %10d visibilities (%.3g%% of MS rows)" % (
                    sel_nrow, sel_nvis, sel_nrow * rpc))
                    if options.channels or options.corrs:
                        print("===>   Chan/corr slicing reduces this to    %12d visibilities (%.3g%% of selection)" % (
                        nvis_A, nvis_A * percent))
                print(
                    "===>   %-29s includes %10d visibilities (%.3g%% of selection)" % (label, nv, nv * percent))
                    
            # now check for stats over CASA flagversions
            flagvers = [tab for tab in glob.glob(f"{msname}.flagversions/flags.*") if Owlcat.tableexists(tab)]
            if flagvers:
                print(f"Stats for CASA-style flagversions:")
                for vers in flagvers:
                    name = os.path.basename(vers).split('.', 1)[1]
                    ftab = Owlcat.table(vers, ack=False)
                    ratio = copy_chunked_flags(ftab, None, options.chunk_size)
                    ftab.close()
                    print(f"    '{name}': {ratio:.2%} of all data flagged.")
            else:
                print("  No CASA-style flagversions available.")
            if options.flag or options.unflag or options.copy or options.fill_legacy or options.remove:
                print("-l/--list was in effect, so all other options were ignored.")
            print("")
            sys.exit(0)

        # else not stats mode, do the actual flagging job
        totrows, sel_nrow, sel_nvis, nvis_A, nvis_B, nvis_C = \
            flagger.xflag(flag=options.flag, unflag=options.unflag, copy=options.copy, fill_legacy=options.fill_legacy,
                          flag_allcorr=options.extend_all_corr,
                          **subset)

        # print stats
        if statonly:
            print("===> No actions were performed. Showing the result of your selection:")
        else:
            print("===> Flagging stats:")
        rpc = 100.0 / totrows if totrows else 0
        print("===>   MS size:               %8d rows" % totrows)
        print("===>   Data/time selection:   %8d rows, %10d visibilities (%.3g%% of MS rows)" % (
        sel_nrow, sel_nvis, sel_nrow * rpc))
        if legacystr:
            print("===>     (over which legacy flags were filled using flagmask %s)" % legacystr)

        percent = 100.0 / sel_nvis if sel_nvis else 0
        if options.channels or options.corrs:
            print("===>   Chan/corr slicing reduces this to     %10d visibilities (%.3g%% of selection)" % (
            nvis_A, nvis_A * percent))
        if not (options.flagmask is None and options.flagmask_all is None and options.flagmask_none is None):
            print("===>   Flag selection reduces this to        %10d visibilities (%.3g%% of selection)" % (
            nvis_B, nvis_B * percent))
        if options.nan or options.above is not None or options.below is not None or \
                options.fm_above is not None or options.fm_below is not None:
            print("===>   Data selection reduces this to         %10d visibilities (%.3g%% of selection)" % (
            nvis_C, nvis_C * percent))
        if unflagstr:
            print("===>     (which were unflagged using flagmask %s)" % unflagstr)
        if flagstr:
            print("===>     (which were flagged using flagmask %s)" % flagstr)

        flagger.close()

    # export flags from file, if so specified
    if options.export:
        try:
            dump = Owlcat.Tables.TableDump(options.export, write=True, compress=True)
            ms = get_ms()
            colnames = set(ms.colnames())
            print("Exporting flags to %s, %d rows:" % (options.export, ms.nrows()))
            for colname in FLAGCOLS:
                if colname in colnames:
                    dump.dump_column(ms, colname, verbose=True)
            ms.close()
            dump.close()
        except:
            traceback.print_exc()
            error("Error exporting flags to %s" % options.export)
        print("Flags exported OK.")

    if options.save:
        flagvers = f"{msname}.flagversions/flags.{options.save}"
        if Owlcat.tableexists(flagvers) and not options.force:
            error(f"Flagversion '{options.save}' already exists. Please run with --force to confirm overwrite.")
        try:
            ms = get_ms(readonly=True)
            fr_desc = ms.getcoldesc("FLAG_ROW")
            fc_desc = ms.getcoldesc("FLAG")
            if os.path.exists(flagvers):
                os.system(f"rm -fr {flagvers}")
            from casacore.tables import maketabdesc
            fr_desc['name'] = "FLAG_ROW"
            fc_desc['name'] = "FLAG"
            td = maketabdesc([dict(name='FLAG_ROW', desc=fr_desc), dict(name='FLAG', desc=fc_desc)])
            ftab = Owlcat.table(flagvers, tabledesc=td, readonly=False, nrow=ms.nrows())
            ratio = copy_chunked_flags(ms, ftab, options.chunk_size)
            ftab.close()
            ms.close()
            print(f"Flag version '{options.save}' saved: {ratio:.2%} flagged.")
            # add to list
            try:
                version_list_file = f"{msname}.flagversions/FLAG_VERSION_LIST"
                version_list = open(version_list_file, encoding="ascii").readlines()
                versions = set(line.split(':', 1)[0].strip() for line in  version_list)
                if options.save in versions:
                    print(f"Flag version '{options.save}' is present in CASA's FLAG_VERSION_LIST.")
                else:
                    with open(version_list_file, "wt", encoding="ascii") as fo:
                        for line in version_list:
                            fo.write(line.rstrip() + "\n")
                        fo.write(f"{options.save} :\n")
                    print(f"Flag version '{options.save}' added to CASA's FLAG_VERSION_LIST.")
            except Exception:
                traceback.print_exc()
                print("There was an error writing CASA's FLAG_VERSION_LIST file.")
                print("Your new flag version might not be recognized by CASA's flagmanager.")

        except:
            traceback.print_exc()
            error(f"Error saving flags to {options.save}, exiting")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup
    from Owlcat.Plotting import MultigridPlot, SkyPlot, PLOT_SINGLE, PLOT_MULTI, PLOT_ERRORBARS

    parser = OptionParser(usage="""%prog: [plots & options] parmtables""",
                          description="""Makes various plots of poitning error solutions.""")

    parser.add_option("-l", "--list", action="store_true",
                      help="lists stuff found in MEP tables, then exits")
    parser.add_option("-c", "--cache", metavar="FILENAME", type="string",
                      help="cache parms to file, which can be fed to plot-de-solutions script")
    parser.add_option("-n", "--nominals", metavar="FILENAME", type="string",
                      help="includes nominal offsets on plot (supply filename)")
    parser.add_option("--nominal-circle", metavar="mDEG", type="float", default=10,
                      help="size of circle around nominal offsets")
    parser.add_option("--circle-radius", metavar="mDEG", type="float", default=0,
                      help="set a fixed size for the circle pointing plot")
    parser.add_option("--from", metavar="TIMESLOT", type="int", default=0,
                      help="extract solutions starting from given timeslot")
    parser.add_option("--to", metavar="TIMESLOT", type="int", default=-1,
                      help="extract solutions until the given timeslot")
    parser.add_option("--antennas", type="string",
                      help="antenna subset, as a comma-separated list of names. Simple wildcards are allowed.")
    parser.add_option("--pt-max", metavar="VALUE", type="float", default=0,
                      help="set fixed plot limits on pointing error vs. time plot")
    parser.add_option("--ft", action="store_true",
                      help="generate plots of pointing offset Fourier components")
    parser.add_option("--ft-max", metavar="VALUE", type="float", default=0,
                      help="set fixed plot limits on Fourier component plot")
    parser.add_option("--wind", metavar="FILE", type="string",
                      help="loads wind data from pickle file, adds wind plots")
    parser.add_option("--prefix", metavar="PREFIX", type="string", default="E::",
                      help="funklet name prefix")

    plotgroup = OptionGroup(parser, "Plotting options")
    outputgroup = OptionGroup(parser, "Output options")
    MultigridPlot.init_options(plotgroup, outputgroup)
    SkyPlot.init_options(plotgroup, outputgroup)

    parser.add_option_group(plotgroup)
    parser.add_option_group(outputgroup)

    (options, args) = parser.parse_args()

    if not args:
        parser.error("No parmtables specified.")

    import os.path
    import os
    import sys

    import numpy
    from Owlcat.ParmTables import ParmTab
    import math

    DEG = math.pi / 180
    ARCMIN = math.pi / (180 * 60)

    SPWS = list(range(len(args)))

    # set of all sources, antennas and correlations
    ANTS = set()

    # complex array of dEs per each src,antenna,corr tuple
    des = {}

    # scan funklet names to build up sets of keys
    oldtable = False
    pt = ParmTab(args[0])
    print(("Table contains:", pt.funklet_names()))
    for name in pt.funklet_names():
        if name.startswith("E::dlm::dl"):
            options.prefix = "E::dlm::"
            oldtable = True
        if name.startswith(options.prefix):
            fields = name.split(':')
            ANTS.add(fields[-1])
    ts_slice = slice(getattr(options, 'from'), options.to if options.to >= 0 else None)
    NTIMES = len(pt.funkset(pt.funklet_names()[0]).get_slice()[ts_slice])
    print(("%d timeslots found in table %s" % (NTIMES, args[0])))

    ANTS = sorted(ANTS)
    if not ANTS:
        print(("No pointing offset solutions found in MEP table %s." % args[0]))
        sys.exit(1)

    if options.list:
        print(("MEP table %s contains pointing offsets for" % args[0]))
        print(("  %d antennas: %s" % (len(ANTS), " ".join(ANTS))))

    # source subset selection
    import fnmatch

    if options.antennas:
        ants = set()
        for a in options.antennas.split(","):
            subset = fnmatch.filter(ANTS, a)
            if subset:
                ants.update(subset)
            else:
                print(("WARNING: \"%s\" does not match any antenna names in MEP table %s." % (a, args[0])))
        ANTS = sorted(ants)
        if not ANTS:
            print("No antennas were selected, check your --antennas option.")
            sys.exit(1)

    print(("Using %d antennas: %s" % (len(ANTS), " ".join(ANTS))))

    interval = None


    def c00(funklet):
        global interval
        if interval is None:
            interval = funklet.domain.time[1] - funklet.domain.time[0]
        if numpy.isscalar(funklet.coeff):
            return funklet.coeff
        else:
            return funklet.coeff.ravel()[0]


    # dl,dm is a 2 x NSPW x  NANT x NTIME array of poitning offsets
    dlm = numpy.zeros((2, len(SPWS), len(ANTS), NTIMES), dtype=float)

    # bsz is a 2 x 2 x NSPW x  NANT x NTIME array of beam sizes
    # first index is x/y, second is l/m
    bsz = numpy.zeros((2, 2, len(SPWS), len(ANTS), NTIMES), dtype=float)
    beam_sizes = 0

    for spw, tabname in enumerate(args):
        print(("Reading", tabname))
        pt = ParmTab(tabname)
        for i, ant in enumerate(ANTS):
            # fill times
            # fill dlm
            if oldtable:
                fsl = pt.funkset('E::dlm::dl:%s' % ant).get_slice()[ts_slice]
                fsm = pt.funkset('E::dlm::dm:%s' % ant).get_slice()[ts_slice]
            else:
                fsl = pt.funkset('%sdl:%s' % (options.prefix, ant)).get_slice()[ts_slice]
                fsm = pt.funkset('%sdm:%s' % (options.prefix, ant)).get_slice()[ts_slice]
            if len(fsl) != len(fsm) or len(fsl) != NTIMES:
                print(("Error: table contains %d funklets for dl and %d for dm; %d expected" % (
                len(fsl), len(fsm), NTIMES)))
                sys.exit(1)
            if not i:
                t0 = numpy.array([funklet.domain.time[0] for funklet in fsl])
                t1 = numpy.array([funklet.domain.time[1] for funklet in fsl])
            dlm[0, spw, i, :] = list(map(c00, fsl))
            dlm[1, spw, i, :] = list(map(c00, fsm))
            # fill beam sizes
            if '%sbeamshape:%s' % (options.prefix, ant) in pt.funklet_names():
                beam_sizes = 1
                fs = pt.funkset('%sbeamshape:%s' % (options.prefix, ant)).get_slice()[ts_slice]
                bsz[0, 0, spw, i, :] = list(map(c00, fs))
            elif '%sbeamshape:xy:lm:%s' % (options.prefix, ant) in pt.funklet_names():
                beam_sizes = 4
                for ixy, xy in enumerate("xy"):
                    for ilm, lm in enumerate("lm"):
                        fs = pt.funkset('%sbeamshape:%s:%s:%s' % (options.prefix, ant, xy, lm)).get_slice()[ts_slice]
                        bsz[ixy, ilm, spw, i, :] = list(map(c00, fs))

    interval = round((interval or 60) / 60)
    print(("Solution interval is", interval, "minutes; total time", (interval * NTIMES) / 60, "hours"))

    # write cache
    if options.cache:
        import pickle

        cachefile = options.cache + '.cache'
        pickle.dump((dlm, bsz, beam_sizes), open(cachefile, 'wb'))
        print(("Cached all structures to file", cachefile))

    if options.wind:
        import pickle

        wind_time, wind_dir, wind_speed = pickle.load(open(options.wind, 'rb'))
        print(("Loaded %d wind samples from %s" % (len(wind_time), options.wind)))
        # get wind data for each interval
        wsp_mean = numpy.zeros(NTIMES, float)
        wsp_std = numpy.zeros(NTIMES, float)
        for itime in range(NTIMES):
            i0 = wind_time.searchsorted(t0[itime])
            i1 = wind_time.searchsorted(t1[itime])
            if i1 - i0 <= 0:
                print(("No wind data for interval %d, cannot plot wind" % itime))
                sys.exit(1)
            wsp_mean[itime] = wind_speed[i0:i1].mean()
            wsp_std[itime] = wind_speed[i0:i1].std()

    # convert dlm to millidegrees
    dlm0 = dlm.copy()
    dlm *= 180 * 1000 / math.pi

    # take mean and std along freq axis
    # these are now 2 x NANT x NTIME arrays
    dlm_mean = dlm.mean(1)
    dlm_std = dlm.std(1)
    bsz_mean = bsz.mean(2)
    bsz_std = bsz.std(2)
    # and along time axis
    # these are now 2 x NSPW x NANT
    dlm_fmean = dlm.mean(3)
    dlm_fstd = dlm.std(3)
    bsz_fmean = bsz.mean(4)
    bsz_fstd = bsz.std(4)

    print(("Read %d parmtables" % len(args)))

    from Owlcat.Plotting import MultigridPlot, PLOT_SINGLE, PLOT_MULTI, PLOT_ERRORBARS, PLOT_BARPLOT

    # initialize plot object
    figplot = MultigridPlot(options)
    make_figure = figplot.make_figure

    skyplot = SkyPlot(options)
    make_skymap = skyplot.make_figure

    funcs = [
        lambda iant: (dlm_mean[0, iant, :], dlm_std[0, iant, :]),
        lambda iant: ("mean %.2f" % numpy.array([dlm_mean[0, iant, :].mean()]),
                      "+/- %.2f" % numpy.array([dlm_mean[0, iant, :].std()])),
        lambda iant: (dlm_mean[1, iant, :], dlm_std[1, iant, :]),
        lambda iant: ("mean %.2f" % numpy.array([dlm_mean[1, iant, :].mean()]),
                      "+/- %.2f" % numpy.array([dlm_mean[1, iant, :].std()])),
        lambda iant: (dlm_fmean[0, :, iant], dlm_fstd[0, :, iant]),
        lambda iant: (dlm_fmean[1, :, iant], dlm_fstd[1, :, iant])
    ]

    labels = ["dl", "", "dm", "", "dl, freq", "dm, freq"]
    make_figure(enumerate(labels), enumerate(ANTS),
                lambda i, iant: funcs[i](iant),
                hline=0, ylock=(-options.pt_max, options.pt_max) if options.pt_max else True, figsize=(290, 150),
                mode=PLOT_ERRORBARS,
                suptitle="Pointing offset mean & stddev across all bands (top two plots) and times (bottom two plots), mdeg",
                save="Epnt_mean")

    if options.wind:
        # work out wind data for each solve interval
        funcs = [
            lambda iant: (dlm_mean[0, iant, :], dlm_std[0, iant, :]),
            lambda iant: (dlm_mean[1, iant, :], dlm_std[1, iant, :]),
            lambda iant: (wsp_mean, wsp_std),
        ]

        labels = ["dl", "dm", "wind, m/s"]
        make_figure(enumerate(labels), enumerate(ANTS),
                    lambda i, iant: funcs[i](iant),
                    hline=0, ylock="row", figsize=(290, 75), mode=PLOT_ERRORBARS,
                    suptitle="Pointing offset mean & stddev, versus wind speed",
                    save="Epnt_wind")

    for iant, ant in enumerate(ANTS):
        print(("mean offset %s: %6.2f %6.2f" % (ant, dlm_mean[0, iant, :].mean(), dlm_mean[1, iant, :].mean())))

    if options.ft:
        import numpy.fft

        # fft scaling is 1/NTIMES, then later we multiply the non-0th-order terms by 2 since each fourier component's
        # power is split between itself and the conjugate
        ft_slice = slice(0, NTIMES / 2 + 1)
        dlm_fft = abs(numpy.fft.fft(dlm_mean))[:, :, ft_slice] / NTIMES
        dlm_fft[:, :, 1:] *= 2
        periods = 1 / abs(numpy.fft.fftfreq(NTIMES, interval)[1:])
        dlm_fftmax = dlm_fft[:, :, 1:].max(2)
        dlm_fftper = numpy.zeros((2, len(ANTS)))
        for i in 0, 1:
            for iant in range(len(ANTS)):
                dlm_fftper[i, iant] = periods[numpy.where(dlm_fft[i, iant, 1:] == dlm_fftmax[i, iant])]
        funcs = [
            lambda iant: dlm_fft[0, iant, :],
            lambda iant: (
            "offset %.1f" % dlm_fft[0, iant, 0], "~max %.1f" % dlm_fftmax[0, iant], "@%d min" % dlm_fftper[0, iant]),
            lambda iant: dlm_fft[1, iant, :],
            lambda iant: (
            "offset %.1f" % dlm_fft[1, iant, 0], "~max %.1f" % dlm_fftmax[1, iant], "@%d min" % dlm_fftper[1, iant]),
        ]
        labels = ["FT(dl)", "", "FT(dm)", ""]
        make_figure(enumerate(labels), enumerate(ANTS),
                    lambda i, iant: funcs[i](iant),
                    hline=0, ylock=(0, options.ft_max or dlm_fft.max()), figsize=(290, 25 * len(labels)),
                    mode=PLOT_BARPLOT,
                    suptitle="Pointing offset Fourier components",
                    save="Epnt_ft")

    if beam_sizes == 4:
        funcs = []
        for i0 in range(2):
            for j0 in range(2):
                funcs += [
                    lambda iant, i=i0, j=j0: (bsz_mean[i, j, iant, :], bsz_std[i, j, iant, :]),
                    lambda iant, i=i0, j=j0: ("mean %.2f" % numpy.array([bsz_mean[i, j, iant, :].mean()]),
                                              "+/- %.2f" % numpy.array([bsz_mean[i, j, iant, :].std()]))
                ]

        make_figure(enumerate(("Lx", "", "Mx", "", "Ly", "", "My", "")), enumerate(ANTS),
                    lambda i, iant: funcs[i](iant),
                    hline=1, ylock=True, figsize=(290, 210), mode=PLOT_ERRORBARS,
                    mean_format="%.4f",
                    suptitle="Beam extent in L/M, for X and Y dipoles, mean over time",
                    save="Eshape_mean")

        funcs = []
        for i0 in range(2):
            for j0 in range(2):
                funcs += [
                    lambda iant, i=i0, j=j0: (bsz_fmean[i, j, :, iant], bsz_fstd[i, j, :, iant]),
                    lambda iant, i=i0, j=j0: ("mean %.2f" % numpy.array([bsz_fmean[i, j, :, iant].mean()]),
                                              "+/- %.2f" % numpy.array([bsz_fmean[i, j, :, iant].std()]))
                ]
        make_figure(enumerate(("Lx", "", "Mx", "", "Ly", "", "My", "")), enumerate(ANTS),
                    lambda i, iant: funcs[i](iant),
                    mean_format="%.4f",
                    hline=1, ylock=True, figsize=(290, 210), mode=PLOT_ERRORBARS,
                    suptitle="Beam extent in L/M, for X/Y dipoles, mean over frequency",
                    save="Eshape_mean_fq")
    elif beam_sizes == 1:
        funcs = [
            lambda iant: (bsz_mean[0, 0, iant, :], bsz_std[0, 0, iant, :]),
            lambda iant: ("mean %.2f" % numpy.array([bsz_mean[0, 0, iant, :].mean()]),
                          "+/- %.2f" % numpy.array([bsz_mean[0, 0, iant, :].std()])),
            lambda iant: (bsz_fmean[0, 0, :, iant], bsz_fstd[0, 0, :, iant]),
        ]
        make_figure(enumerate(("size", "", "size fq")), enumerate(ANTS),
                    lambda i, iant: funcs[i](iant),
                    mean_format="%.4f",
                    hline=1, ylock=True, figsize=(290, 75), mode=PLOT_ERRORBARS,
                    suptitle="Beam extent",
                    save="Eshape")

    # make skymap with average pointings
    ll = []
    mm = []
    markers = []

    # add nominal mispointings
    if options.nominals:
        exec(open(options.nominals))
    else:
        nominals = {}

    circle_radius = options.circle_radius / (1000 / 60.)
    nominal_radius = options.nominal_circle / (1000 / 60.)

    circlex1 = numpy.cos(numpy.arange(0, 1.05, .05) * math.pi * 2)
    circley1 = numpy.sin(numpy.arange(0, 1.05, .05) * math.pi * 2)
    circlex = circlex1 * nominal_radius
    circley = circley1 * nominal_radius

    ll.append(0)
    mm.append(0)
    markers.append(
        ("plot", (circlex, circley, ":"),
         dict(color='blue'))
    )

    for iant, ant in enumerate(ANTS):
        dl = dlm0[0, :, iant, :]
        dm = dlm0[1, :, iant, :]
        dl_mean = dl.mean()
        dl_std = dl.std()
        dm_mean = dm.mean()
        dm_std = dm.std()
        color = "pink" if nominals else "blue"
        # plot nominal position, if available
        if ant in nominals:
            dl0, dm0 = nominals[ant]
            markers += [
                ("text", (dl0 / ARCMIN, dm0 / ARCMIN, ant),
                 dict(color='blue', ha='center', va='center', size='large', weight='bold')),
                ("plot", (dl0 / ARCMIN + circlex, dm0 / ARCMIN + circley, ":"),
                 dict(color='blue')),
                ("plot", ((dl0 / ARCMIN, dl_mean / ARCMIN), (dm0 / ARCMIN, dm_mean / ARCMIN), ':'),
                 dict(color='grey')),
            ]
            ll += [dl0, dl0, dl0]
            mm += [dm0, dl0, dl0]
            color = "red"
        # plot fitted position
        markers += [
            ("errorbar", (dl_mean / ARCMIN, dm_mean / ARCMIN, dm_std / ARCMIN, dl_std / ARCMIN), dict(color='#A0A0A0')),
            ("text", (dl_mean / ARCMIN, dm_mean / ARCMIN, "%s " % ant),
             dict(color=color, ha='center', va='center', size='large', weight='bold',
                  bbox=dict(fc='white', ec='none' if beam_sizes else 'grey'))),
        ]
        ll += [dl_mean, dl_mean]
        mm += [dm_mean, dm_mean]
        # plot beam size
        if beam_sizes == 1:
            bsz = bsz_mean[0, 0, iant, :].mean()
            bsz = min(1.2, max(bsz, .8))
            bsz = bsz - 1
            radius0 = nominal_radius / 4
            radius = radius0 + bsz * 5 * radius0
            markers += [("plot",
                         (dl_mean / ARCMIN + circlex1 * radius0,
                          dm_mean / ARCMIN + circley1 * radius0, '-'),
                         dict(color='grey', zorder=10, alpha=0.5))]
            markers += [("plot",
                         (dl_mean / ARCMIN + circlex1 * radius,
                          dm_mean / ARCMIN + circley1 * radius, '-'),
                         dict(color='green' if bsz > .01 else 'purple' if bsz < -.01 else 'grey', zorder=20,
                              alpha=0.5))]
            ll += [dl_mean, dl_mean]
            mm += [dm_mean, dm_mean]

    make_skymap(numpy.array(ll), numpy.array(mm), markers,
                zero_lines=False, radius=circle_radius or None,
                suptitle="Fitted pointing offsets", save="Eplot")

    if options.output_type.upper() == "X11":
        from pylab import plt

        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import sys
import traceback
import gzip
import pickle
import Owlcat

flagger = parser = ms = msname = None


def error(message):
    print(("%s: %s" % (os.path.basename(sys.argv[0]), message)))
    sys.exit(1)


def get_ms():
    global ms
    global msname
    if not ms:
        ms = Owlcat.table(msname)
    return ms


if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup

    parser = OptionParser(usage="""%prog: [actions] [options] MS COLUMN FILENAME[.gz]""",
                          description="Exports MS column to an external file, which can be reloaded with import-ms-column."
                          )
    parser.add_option("-r", "--row-step", type="int",
                      help="how many rows to step over at a time, default is %default")
    #  parser.add_option("-z","--gzip",action="store_true",
    #                    help="compress data with gzip")
    #  parser.add_option("-v","--verbose",metavar="LEVEL",type="int",
    #                    help="verbosity level for messages. Higher is more verbose, default is 0.")
    parser.set_defaults(verbose=0, row_step=200000)

    # parse args
    (options, args) = parser.parse_args()
    if len(args) != 3:
        parser.error("Incorrect number of arguments. Use '-h' for help.")
    msname, colname, filename = args

    try:
        gzf = gzip.GzipFile(filename, "wb") if filename.endswith(".gz") else open(filename, "wb")
        # write stuff
        ms = get_ms()
        print(("Opened MS %s" % msname))
        if colname not in ms.colnames():
            error("Column %s not found" % colname)
        nrows = ms.nrows()
        print(("Exporting %s to %s:" % (colname, filename)))
        pickle.dump((options.row_step, nrows), gzf)
        for r0 in range(0, nrows, options.row_step):
            print(("%d/%d" % (r0, nrows)))
            col = ms.getcol(colname, r0, options.row_step)
            pickle.dump(col, gzf)
        ms.close()
        gzf.close()
    except:
        traceback.print_exc()
        error("Error exporting %s to %s" % (colname, filename))
    print("Column exported OK.")



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import sys
import logging
import os

import casacore.measures
import casacore.quanta as qa
import numpy as np

from casacore.tables import table
from collections import namedtuple

from bokeh.plotting import figure
from bokeh.layouts import column
from bokeh.models import ColumnDataSource, PreText
from bokeh.io import output_file, save

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)-15s %(filename)s %(levelname)s %(message)s")

logging.disable(logging.CRITICAL)


def wgs84_to_ecef(lon, lat, alt):
    """
    Convert wgs84(latitude (deg), longitude(deg), elevation(deg)) to
    Earth Centred Earth fixed coordinates (X(m), Y(m), Z(m)).

    Coordinates in the ITRF format should be converted to WGS84 coordinate system for consistency.
    This function is an implementation based on the following reference:

    https://docs.hisparc.nl/coordinates/HiSPARC_coordinates.pdf

    Parameters
    ----------:
    lat: :obj:`float`
        Latitude in radians
    lon: :obj:`float`
        Longitude in radians
    alt: :obj:`float`
        Altitude in metres

    Returns
    -------
    X, Y, Z:  ECEF coordinates in metres
    """

    logging.debug("Converting WGS84 to ECF")

    # set up earth's shape ellipsoid approximation
    # semi major axis
    a = 6378137.0

    # flattening
    f = 1 / 298.257223563

    # semi-minor axis
    b = a - a * f

    # eccentricity
    e = np.sqrt((2 * f) - (f**2))

    # Normal: Distance between a location on the ellipsoid and the
    # intersection of its nromal and the ellipsoid's z-axis
    N = a / np.sqrt(1 - e**2 * np.sin(lat)**2)

    # altitude
    h = alt

    X = (N + h) * np.cos(lat) * np.cos(lon)
    Y = (N + h) * np.cos(lat) * np.sin(lon)
    Z = (((b**2 / a**2) * N) + h) * np.sin(lat)

    return X, Y, Z


def ecef_to_enu(x, y, z, r_lon, r_lat, r_alt):
    """
    Convert ECEF coordinates to East, North, Up (m). Because the ENU is a
    local coordinate system (to the observer/ antenna in this case)
    an reference position / origin is required from which calculate the
    relative positions.

    Parameters
    ----------
    x: :obj:`float`
        X in metres
    y: :obj:`float`
        Y in metres
    z: :obj:`float`
        Z in metres
    r_lon: :obj:`float`
        Reference longitude in radians
    r_lat: :obj:`float`
        Reference latitude in radians
    r_alt: :obj:`float`
        Reference altitude in metres

    Returns
    -------
    ee, en, eu: :obj:`float`
        East, North Up positions in metres
    """

    logging.debug("Converting ECEF to ENU")

    # convert reference coordinates to ECEF
    rx, ry, rz = wgs84_to_ecef(r_lon, r_lat, r_alt)

    dx = x - rx
    dy = y - ry
    dz = z - rz

    ee = (-np.sin(r_lon) * dx) \
        + (np.cos(r_lon) * dy) \
        + (0 * dz)

    en = (-np.sin(r_lat) * np.cos(r_lon) * dx) \
        + (-np.sin(r_lat) * np.sin(r_lon) * dy) \
        + (np.cos(r_lat) * dz)

    eu = (np.cos(r_lat) * np.cos(r_lon) * dx) \
        + (np.cos(r_lat) * np.sin(r_lon) * dy) \
        + (np.sin(r_lat) * dz)

    return ee, en, eu


def wgs84_to_enu(lon, lat, alt, r_lon, r_lat, r_alt):
    """Convert WSG84 / ITRF to ENU"""
    X, Y, Z = wgs84_to_ecef(lon, lat, alt)
    e, n, u = ecef_to_enu(X, Y, Z, r_lon, r_lat, r_alt)

    return e, n, u


def get_antenna_data(ms_name):
    """Return antenna ITRF positions and other data"""

    logging.debug("Getting antenna data")

    with table(ms_name, ack=False) as ms: 
        if "OBSERVATION" in ms.keywordnames():
            with table("::".join((ms_name, "OBSERVATION")), ack=False) as observ_sub:
                obs = observ_sub.getcell("TELESCOPE_NAME", 0)
        else:
            obs = "unknown"

        if "ANTENNA" in ms.keywordnames():
            ant_tab = ms.getkeyword("ANTENNA").split()[-1]
        else:
            ant_tab = ms_name

        with table(ant_tab, ack=False) as ant_sub:
            names = ant_sub.getcol("NAME")
            stations = ant_sub.getcol("STATION")
            offsets = ant_sub.getcol("OFFSET")
            positions = ant_sub.getcol("POSITION")
            indices = ant_sub.rownumbers()

    logging.debug(f"Telescope name: {obs}")
    logging.debug(f"Found: {len(names)} antennas")

    meta = namedtuple("antenna",
                      "names, stations, positions, offsets, telescope, indices")
    antenna = meta(names=names, stations=stations, positions=positions,
                   offsets=offsets, telescope=obs, indices=indices)

    return antenna


def get_antenna_coords(positions, obs, offsets=None, cofa=None):
    """Return antenna COFA coordinates, and antenna coordinates

    Parameters
    ----------
    positions: :obj:`np.array`
        Antenna ITRF positions
    offset: :obj:`np.array`
        Antenna offset positions
    obs: :obj:`str`
        Telescope name or observatory name
    """

    logging.debug("Getting antenna coordinates")

    me = casacore.measures.measures()

    quantify = lambda coord, unit: qa.quantity(coord, unit)

    # make ITRF positions casacore quantities so that we can process using
    # casacore measurements

    if casacore.measures.is_measure(positions):
        baseline_me = positions

    else:
        xq, yq, zq = list(map(quantify, positions.T, ["m", "m", "m"]))

        if offsets is not None:
            offsets = list(map(quantify, offsets.T, ["m", "m", "m"]))

        # make antenna positions into casacore measurements
        baseline_me = me.position('itrf', v0=xq, v1=yq, v2=zq,
                                  off=offsets)

    if cofa:
        if casacore.measures.is_measure(cofa):
            obs_cofa = cofa
        else:
            cofa = list(map(quantify, [float(_) for _ in cofa.split(",")],
                            ["deg", "deg", "m"]))
            obs_cofa = me.position("WGS84", *cofa)
    else:
        # get the list of observatories
        obs_list = me.get_observatories()

        if obs in obs_list:
            # Get a (position) measure Centre of Array (cofa) for the given
            # osbervatory. This is in WSG84 reference frame.
            obs_cofa = me.observatory(obs)
        else:
            # first antenna position
            i_ant = positions[0]
            ixq, iyq, izq = list(map(quantify, i_ant, ["m", "m", "m"]))
            obs_cofa = me.position('itrf', v0=ixq, v1=iyq, v2=izq,
                                   off=offsets)
            logging.debug(f"Centre of Array of Observatory: {obs} not found.")
            logging.debug("Defaulting to first antenna of the array")

    # get the cofa's reference frame
    obs_cofa = me.measure(obs_cofa, "WGS84")

    logging.debug("COFA: " +
                  ", ".join([_.formatted("dd.mm.ss.t..")
                             for _ in me.get_value(obs_cofa)]))

    # make this position the reference frame of the entire measurement
    me.doframe(obs_cofa)

    # latitude, longitude and elevation in degree, minute second string form
    # get this before changing reference frame
    str_lon_lat_el = me.get_value(baseline_me)
    str_lon_lat_el = [_.formatted("dd.mm.ss.t..").strip(
        "[]").split(",") for _ in str_lon_lat_el]

    # change the baseline measurement to WGS84
    baseline_me = me.measure(baseline_me, "WGS84")

    # get the wsg84 coordinates in radians for the COFA
    cofa_coords = me.get_value(obs_cofa)
    cofa_coords = [_.get_value() for _ in cofa_coords]

    # for the antennas
    bl_coords = me.get_value(baseline_me)
    bl_coords = [_.get_value() for _ in bl_coords]

    rf_coords = namedtuple("coords", "cofa, ant_rad, ant_str")
    coords = rf_coords(cofa=cofa_coords, ant_rad=bl_coords,
                       ant_str=str_lon_lat_el)

    return coords


def read_underlay_file(in_file):
    """Read file containing complete antenna layout for plotting.

    An underlay file is a JSON file containing the coordinates of all of a
    telescope's antennas. This includes the antennas that were not used for
    the observation. By default, this script will only plot the antennas that
    were used in the observation i.e. in the current MS. If the underlay file
    is provided, antennas used in the observation will be shown in colour, in
    -30.54417addition to the rest that were inactive, which will be denoted
    by the colour grey.

    Parameters
    ----------
    in_file: :obj:`str`
        Input JSON file containing the telescope's antenna coordinates
    """
    # allowed units
    allowed_us = {"deg", "m", "rad"}

    # allowed rfs
    allowed_rfs = {"WGS84", "ITRF"}

    with open(in_file, "r") as rf:
        data = json.load(rf)

    for spec in ["cofa", "antenna"]:
        rf = [data[spec]["rf"]]
        unit = data[spec]["units"]

        if not set(rf) <= allowed_rfs:
            print(f"Invalid rf {rf} for {spec}")
            sys.exit(-1)

        if not set(unit) <= allowed_us:
            print(f"Invalid unit {unit} for {spec}")
            sys.exit(-1)

    return data


def get_underlay_data(in_file):
    """Parse underlay file data and create position measures for the
    Centre of Array (COFA) and the antenna coordinates
    """
    data = read_underlay_file(in_file)

    obs = data["obs"]

    cofa_units = data["cofa"]["units"]
    cofa_rf = data["cofa"]["rf"]
    cofa_coords = data["cofa"]["coords"]

    ant_units = data["antenna"]["units"]
    ant_rf = data["antenna"]["rf"]
    ant_pos = data["antenna"]["coords"]
    names = list(ant_pos.keys())
    ant_coords = np.array(list(ant_pos.values()))

    quantify = lambda coord, unit: qa.quantity(coord, unit)
    me = casacore.measures.measures()

    c_lon, c_lat, c_el = list(map(quantify, cofa_coords, cofa_units))
    cofa_me = me.position(rf=cofa_rf, v0=c_lon, v1=c_lat, v2=c_el)

    a_lon, a_lat, a_el = list(map(quantify, ant_coords.T, ant_units))
    ant_me = me.position(rf=ant_rf, v0=a_lon, v1=a_lat, v2=a_el)

    underlay = namedtuple("underlay",
                          "cofa, names, positions, offsets, telescope")
    u_data = underlay(names=names, positions=ant_me, cofa=cofa_me,
                      offsets=None, telescope=obs)

    return u_data


def plot_antennas(source, ms_name):
    """Make bokeh plot of the antennas

    Parameters
    ----------
    source: :obj:`ColumnDataSource`
        Data source containing the plot's data
    ms_name: :obj:`str`
        Name of the current MS

    Return
    ------
    out_layout: Bokeh layout object containing the plots
    """

    tooltips = [
        ("Antenna", "@name"),
        ("Station", "@station"),
        ("Index in MS", "@indices"),
        ("Index in full array", "@rindices"),
        ("Lon, Lat [dms]", "@lon, @lat"),
        ("E,N,U [m]", "(@e{0.000}, @n{0.000}, @u{0.000})"),
        ("ITRF (x, y, z) [m]", "(@x{0.000}, @y{0.000}, @z{0.000})"),
    ]

    logging.debug("Initialising figure")

    nants = len(source["name"])
    obs = source.pop("obs")
    ranges = None

    fig = figure(plot_width=720, plot_height=720, sizing_mode="scale_height",
                 x_axis_label="East [m]", y_axis_label="North [m]",
                 title="Offset From Array Centre", toolbar_location="above",
                 tooltips=tooltips, x_range=ranges, y_range=ranges)

    fig.update(outline_line_width=2, outline_line_color="#017afe",
               outline_line_alpha=0.4)

    fig.title.update(align="center", text_font_size="15pt")

    fig.xaxis.update(axis_label_text_font="monospace",
                     axis_label_text_font_size="10pt",
                     axis_label_text_font_style="normal")
    fig.yaxis.update(axis_label_text_font="monospace",
                     axis_label_text_font_size="10pt",
                     axis_label_text_font_style="normal")

    color = source.pop("color")

    # create reference data source object
    r_source = {}
    for refs in ["re", "rn", "ru", "rname"]:
        r_source[refs] = source.pop(refs)

    r_source = ColumnDataSource(data=r_source)

    source = ColumnDataSource(data=source)

    fig.circle(x="re", y="rn", source=r_source, color="grey", fill_color=None,
               size=30)
    p = fig.scatter("e", "n", line_width=2, line_color=color, marker="circle",
                    fill_color=None, size=30, source=source)
    fig.text(x="re", y="rn", text="rname", text_align="center",
             text_baseline="middle", text_font_size="7pt", source=r_source)

    fig.hover.renderers = [p]

    infos = f"MS       : {ms_name}\n"
    infos += f"Telescope: {obs}\n"
    infos += f"Antennas : {nants}"
    pre = PreText(text=infos, sizing_mode="stretch_width")

    out_layout = column(children=[pre, fig], sizing_mode="stretch_both")

    return out_layout


def main(args):

    ms_name = args.ms_name

    if args.o_file:
        o_file = args.o_file
    else:
        o_file = os.path.basename(os.path.abspath(ms_name))
        o_file += "_antennas.html"

    output_file(o_file)

    # get colors or set default active colour
    if args.cmap:
        cmap = args.cmap
    else:
        cmap = "#FC8103"

    # get information from antenna data table
    ant = get_antenna_data(ms_name)

    # gather antenna coordinates
    coords = get_antenna_coords(ant.positions, ant.telescope,
                                offsets=ant.offsets, cofa=args.cofa)

    # antenna coords in lon/ lat in string format
    str_lon, str_lat, str_el = coords.ant_str

    # ITRF positions
    x, y, z = ant.positions.T.tolist()

    e, n, u = wgs84_to_enu(*coords.ant_rad, *coords.cofa)

    if args.u_file:
        u_file = args.u_file
    else:
        if ant.telescope == "MeerKAT":
            u_file = os.path.join(os.path.dirname(__file__),
                                  "data",
                                  "MeerKAT_WGS84_underlay.json")
        else:
            u_file = None

    # get underlay data
    if u_file:
        u_data = get_underlay_data(u_file)
        u_coords = get_antenna_coords(
            u_data.positions, u_data.telescope, offsets=u_data.offsets,
            cofa=u_data.cofa)

        re, rn, ru = wgs84_to_enu(*u_coords.ant_rad, *u_coords.cofa)
        r_names = u_data.names
    else:
        re, rn, ru = e, n, u
        r_names = ant.names

    source = dict(e=e, n=n, u=u, lon=str_lon, lat=str_lat, el=str_el,
                  x=x, y=y, z=z, indices=ant.indices, color=cmap,
                  name=ant.names, station=ant.stations, obs=ant.telescope,
                  re=re, rn=rn, ru=ru, rname=r_names,
                  rindices=np.where(np.isin(r_names, ant.names))[0])

    plot = plot_antennas(source, ms_name)

    save(plot)

    print(f"Antenna plot at: {o_file}")



# -*- coding: utf-8 -*-

import time
import os.path
import sys
import traceback
import pickle
import copy
import numpy
import numpy.ma
from past.builtins import cmp

from Timba import dmi
from Timba import mequtils
from Timba.parmtables import FastParmTable
from Timba.Meq import meq

# debug printing
import Kittens.utils

verbosity = Kittens.utils.verbosity(name="ParmTables")
dprint = verbosity.dprint
dprintf = verbosity.dprintf


def verbose(level):
    """Changes the verbosity level"""
    verbosity.set_verbose(level)


def _int_or_str(x):
    """helper function, converts string to integer if possible"""
    try:
        return int(x)
    except:
        return x


def cmp_qualified_names(a, b):
    """compares two qualified names: splits them at ':', and comparing the subfields in a numeric sense"""
    from past.builtins import cmp
    return cmp(list(map(_int_or_str, a.split(':'))), list(map(_int_or_str, b.split(':'))))


def sort_qualified_names(inlist):
    """Sorts qualified names as follows: splits into fields at ':', and sorts according to fields.
    If field is a numbers, sorts in the numeric sense."""
    from functools import cmp_to_key
    return sorted(inlist, key=cmp_to_key(cmp_qualified_names))


class _AxisStats(object):
    """_AxisStats represents information about one axis in the parmtable. It is created internally
    by ParmTab."""

    def __init__(self, name):
        self.name = str(name).lower()
        self.cells = {}

    def empty(self):
        return not self.cells

    def add_cell(self, x1, x2):
        x0 = (x1 + x2) / 2
        self.cells[x0] = max(x2 - x1, self.cells.get(x0, 0))

    def update(self):
        if self.cells:
            # axis range
            self.minmax = min([x0 - dx / 2 for x0, dx in list(self.cells.items())]), max(
                [x0 + dx / 2 for x0, dx in list(self.cells.items())])
            # grid is simply a sorted list of cell values
            self.grid = list(self.cells.keys())
            self.grid.sort()
            # cell_index gives the orfinal number of each cell in the grid
            self.cell_index = dict([(x0, i) for i, x0 in enumerate(self.grid)])

    def lookup_cell(self, x1, x2):
        return self.cell_index[(x1 + x2) / 2]


class DomainSlicing(list):
    """A DomainSlicing represents a slicing of the ParmTable domain.
    Basically, it is a list of slice_index tuples, with each tuple corresponding to a slice through
    the domain.
    """

    def __init__(self, slice_axes, domain_subset):
        """The constructor makes a slicing from a list of axes to be incorporated in the slice.
        All other axes will be iterated over."""
        list.__init__(self)
        self.append([])
        for iaxis, axis_subset in enumerate(domain_subset):
            if axis_subset is None or iaxis in slice_axes or mequtils.get_axis_id(iaxis) in slice_axes:
                self[:] = [sl + [None] for sl in self]
            else:
                self[:] = [sl + [i] for sl in self for i in axis_subset]


class FunkSlice(object):
    """FunkSlice represents a slice of funklets from a parmtable."""

    def __init__(self, parmtab, name, funklist, index, iaxes, axes=None):
        self.pt = parmtab
        self.name = name
        self.funklets = funklist
        self.slice_index = index
        self.slice_iaxes = iaxes
        self.slice_axes = axes or list(map(mequtils.get_axis_id, iaxes))
        self.rank = len(iaxes)

    def __len__(self):
        return len(self.funklets)

    def __getitem__(self, key):
        return self.funklets[key]

    def __iter__(self):
        return iter(self.funklets)

    def array(self, coeff=0, fill_value=0, masked=True, collapse=True):
        """Returns funklet coefficients arranged into a hypercube.
        'coeff' is applied as an index into each funklet's coeff array, so coeff=0 or coeff=(0,0) selects
        c00, coeff=(0,1) selects c01, etc. IndexError is raised when a non-existing coeff is selected.
        The return value is an array with the same dimensions as the slice. If funklets for some
        domains are missing, the missing values will be set to fill_value. If masked=True, the return value
        will also be a masked array, with missing values masked. If masked=False, return value
        if an ordinary array.
        If 'collapse' is False, axes of the array will have a one-to-one correspondence with domain axes
            (so e.g. a slice for time=0,freq=0,l=*,m=* will have shape [1,1,nl,nm]
        If 'collapse' is True, axes not in this slice will be eliminated.
            (so e.g. the same slice will have shape [nl,nm]
        """
        # convert things like (0,0) into None
        if not coeff or not any(coeff):
            coeff = None
            # initially array is of uncollapsed: one axis for each dimension (up to max_axies), with those not in the
        # slice having a size of 1. Extra axrs will be trimmed later.
        shape = [1] * mequtils.max_axis
        for iaxis in self.slice_iaxes:
            shape[iaxis] = len(self.pt.axis_stats(iaxis).grid)
        # init empty arrays. Mask is all True initially, cleared as filled
        arr = numpy.zeros(shape, float)
        arr[...] = fill_value
        mask = numpy.ones(shape, bool)
        # index object: modified and used below
        idx0 = [0] * mequtils.max_axis
        # now go through funklets and fill them in
        for funk in self.funklets:
            if numpy.isscalar(funk.coeff):
                if coeff is not None:
                    raise IndexError("invalid coeff index %s (funklet is scalar)" % coeff)
                val = funk.coeff
            else:
                try:
                    val = funk.coeff[coeff]
                except:
                    raise IndexError("invalid coeff index %s (funklet coeffs are %s)" % (coeff, funk.coeff.shape))
            # funk.slice_index is the global index of this funklet. We want to use just the axes
            # found in our slice for assigning to the array
            for iaxis in self.slice_iaxes:
                idx0[iaxis] = funk.slice_index[iaxis]
            idx = tuple(idx0)
            try:
                arr[idx] = val
                mask[idx] = False
            except:
                raise
        # make masked array if needed
        if masked:
            arr = numpy.ma.masked_array(arr, mask, fill_value=fill_value)
        # reshape
        if collapse:
            arr.shape = [shape[iaxis] for iaxis in self.slice_iaxes]
        else:
            arr.shape = shape[:(self.slice_iaxes[-1] + 1)]
        return arr


class FunkSet(object):
    """FunkSet represents a set of funklets with the same name"""

    def __init__(self, parmtab, name):
        self.pt = parmtab
        self.name = name

    def get_slice(self, *index, **axes):
        """get_slice() returns a FunkSlice of all funklets matching a specified slice.
        Slice can be specified in one of two ways: by a vector of axis indices, or by keywords.
        E.g.:
          get_slice(None,0) or get_slice(freq=0)
        returns all funklets for 'name' and the first frequency (first form assumes freq is axis 1,
        so time=None and freq=0)
          get_slice(0) or get_slice(time=0)
        returns all funklets for 'name' and the first timeslot (first form assumes time is axis 0),
          get_slice(1,2) or get_slice(time=1,freq=2)
        returns all funklets for 'name', timeslot 1, frequency 2.
        """
        # make sure index is a vector of max_axes length
        if len(index) < mequtils.max_axis:
            index = list(index) + [None] * (mequtils.max_axis - len(index))
        else:
            index = list(index)
        # set additional indices from keywords
        for axis, num in list(axes.items()):
            index[mequtils.get_axis_number(axis)] = num
        # build up list of full indices corresponding to specified slice
        slice_iaxis = []
        indices = [[]]
        for iaxis, axis_idx in enumerate(index):
            stats = self.pt.axis_stats(iaxis)
            # for empty axes, or axes specifed in our call, append number to all indices as is
            if axis_idx is not None or stats.empty():
                list([idx.append(axis_idx) for idx in indices])
            # non-empty axes NOT specified in our call will be part of the slice
            else:
                slice_iaxis.append(iaxis)
                indices = [idx + [i] for idx in indices for i in range(len(stats.grid))]
        # now make funklet list
        funkslice = FunkSlice(self.pt, self.name, [], index, slice_iaxis)
        for idx in indices:
            idx = tuple(idx)
            idom = self.pt._domain_reverse_index.get(idx, None)
            if idom is not None:
                funk = self.pt.parmtable().get_funklet(self.name, idom)
                if funk:
                    funk.domain_index = idom
                    funk.slice_index = idx
                    funkslice.funklets.append(funk)
        return funkslice

    def __call__(self, *index, **axes):
        """The () operator on a FunkSet is equivalent to get_slice()"""
        return self.get_slice(*index, **axes)

    def array(self, coeff=0, fill_value=0, masked=True, collapse=True):
        """Makes array corresponding to whole FunkSet. This function will also maintain a disk cache
        of the array, and read it in or regenerate it as needed (unlike FunkSlice.array(), which
        always builds its arrays from scratch.)
        \n\n""" + FunkSlice.array.__doc__
        # see if we have a cached array
        cachefile = os.path.join(self.pt.filename, "array.%s.%s.cache" % (self.name, coeff))
        arr = None
        if os.path.exists(cachefile) and os.path.getmtime(cachefile) >= self.pt.mtime:
            try:
                arr = pickle.load(open(cachefile, 'rb'))
                dprintf(2, "read cache %s\n" % cachefile)
            except:
                dprintf(0, "error reading cached array %s, will regenerate\n" % cachefile)
                arr = None
        # regenerate array if not read
        if arr is None:
            dprintf(2, "filling array for %s.%s\n", self.name, coeff)
            fullslice = self.get_slice()
            # generate full, uncollapsed masked array
            arr = fullslice.array(coeff, fill_value=0, masked=True, collapse=False)
            # write to cache
            try:
                pickle.dump(arr, open(cachefile, "wb"))
            except:
                if verbosity.get_verbose() is not None and verbosity.get_verbose() > 0:
                    traceback.print_exc()
                dprintf(0, "error writing cache array %s, but proceeding anyway\n" % cachefile)
        # now apply the masked and fill_value properties
        if not masked:
            arr = arr.filled(fill_value)
        else:
            arr.fill_value = fill_value
        # and collapse axes if asked
        if collapse:
            arr.shape = [n for i, n in enumerate(arr.shape) if not self.pt.axis_stats(i).empty()]
        return arr

    def apply(self, op_func, slicing=[], outtab=None, remove=False):
        """For each funklet in the subset, takes all funklets along the designated slicing axis
        (i.e. for each slice along the non-listed axes), creates a FunkSlice, and calls
        op_func(slice).
        The return value of op_func() should be either None if no operation was performed, or a list of
        funklets to be written to the output table. This list may also contain strings, which are
        interpreted as funklet names. The default name is the same as the current FunkSet name; a string at
        any position in the funklet list applies to subsequent funklets.
        if 'outtab' is None, a new output table is created. Otherwise set outtab to a filename, or a
        ParmTab, or a FastParmTable.
        If 'remove' is True, input funklets will be removed if an output funklet is returned.
        """
        # resolve tables
        outtab = self.pt.resolve_output_table(outtab)
        pt = self.pt.parmtable()
        # resolve slicing
        slicing = self.pt.make_slicing(slicing)
        # loop over funklets and slices
        num_infunk = num_outfunk = num_slices = 0
        for sl0 in slicing:
            funklets = self.get_slice(*sl0)
            # call reduction function if we find any
            if funklets:
                num_slices += 1
                num_infunk += len(funklets)
                dprintf(4, "%s slice %s: %d funklets found\n", self.name, sl0, len(funklets))
                dprint(6, "first funklet:", funklets[0])
                dprint(6, "last funklet:", funklets[-1])
                try:
                    outfunk = op_func(funklets)
                    dprint(6, "output funklets:", outfunk)
                except:
                    if verbosity.get_verbose() > 0:
                        dprintf(1, "exception computing funklets for %s slice %s\n", self.name, sl0)
                        traceback.print_exc()
                        dprintf(1, "this slice will be ignored\n")
                    outfunk = None
                # if a new funklet was generated, write it to output
                if outfunk:
                    # remove input funklets if successful
                    if remove:
                        dprintf(4, "%s slice %s: removing %d input funklets\n", self.name, sl0, len(funklist))
                        for funk in funklist:
                            try:
                                self.mtime = time.time()
                                self.parmtable(True).delete_funklet(self.name, funk.domain_index)
                            except:
                                if verbosity.get_verbose() > 0:
                                    traceback.print_exc()
                                dprintf(0, "error deleting funklet for %s slice %s\n", self.name, funk.slice_index)
                    dprintf(4, "%s slice %s: writing %d output funklets\n", self.name, sl0, len(outfunk))
                    name = self.name
                    for ff in outfunk:
                        if isinstance(ff, str):
                            name = ff
                        else:
                            num_outfunk += 1
                            try:
                                outtab.mtime = time.time()
                                outtab.parmtable(True).put_funklet(name, ff)
                            except:
                                dprintf(0, "error saving funklet for %s slice %s\n", self.name, sl0)
                                if verbosity.get_verbose() > 0:
                                    traceback.print_exc()
                                dprintf(0, "this slice will be ignored\n")
                                funk = None
                                break
        dprintf(3, "%s: %s() transformed %d input funklets over %d slices into %d output funklets\n", self.name,
                op_func.__name__, num_infunk, num_slices, num_outfunk)


class ParmTab(object):
    """A ParmTab is a wrapper around a FastParmTable providing high-level facilities
     for dealing with funklets."""

    def __init__(self, filename, write=False, new=False):
        """ParmTab constructor.
         'filename' is a ParmTable name, this will be created if it doesn't exist.
         'write' is True to open the table for writing (can always pass False, and re-open for writing later)
         'new' is True to delete any existing parmtable by the same name and open a new one.
        """
        self._pt = self._pt_write = None
        self.load(filename, write=write, new=new)

    def _start_progress(self, label, maxval=100):
        self._progress_label, self._progress_maxval = label, maxval
        sys.stderr.write("%s: 00%%\r" % label)

    def _report_progress(self, value, maxval=None):
        maxval = maxval if maxval is not None else self._progress_maxval
        sys.stderr.write("%s: %-40s\r" % (self._progress_label, "%02d%%" % (100 * value // maxval)))

    def _end_progress(self, keep=True):
        if keep:
            self._report_progress(self._progress_maxval)
            sys.stderr.write("\n")
        else:
            sys.stderr.write("\r%80s\n" % "")

    def load(self, filename, write=False, new=False):
        """Loads the specified parmtable
         'write' is True to open the table for writing (can always pass False, and re-open for writing later)
         'new' is True to delete any existing parmtable by the same name and open a new one.
        """
        if new:
            if os.path.exists(filename):
                dprintf(1, "removing old table %s\n", filename)
                os.system("/bin/rm -fr %s" % filename)
            dprintf(1, "creating new table %s\n", filename)
            write = True
        else:
            dprintf(1, "loading table %s (write=%d)\n", filename, write)
        self.filename = filename
        self.parmtable(write)
        self._cachepath = os.path.join(filename, 'ParmTab.cache')
        self._make_axis_index()

    def merge(self, filename):
        """Merges in the specified parmtable"""
        pt1 = FastParmTable(filename)
        t0 = time.time()
        dprintf(2, "reading funklet list from %s\n", filename)
        funklist = pt1.funklet_list()
        dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
        t0 = time.time()
        nfunk = len(funklist)
        dprintf(1, "merging in %d funklets from table %s\n", nfunk, filename)
        if funklist:
            pt = self.parmtable(True)
            self._start_progress("merging in parmtable %s" % filename, nfunk)
            try:
                for ifunk, (name, idom, domain) in enumerate(funklist):
                    self._report_progress(ifunk)
                    self.mtime = time.time()
                    pt.put_funklet(name, pt1.get_funklet(name, idom))
                dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
                t0 = time.time()
                pt1 = None
                self._make_axis_index()
            finally:
                self._end_progress()

    def close(self):
        """Detaches from FastParmTable object (needed when interacting with a meqserver). It is usually a good idea
        to detach from a aprmtable when not using it for a while (i.e. only reattach for the duration of a lengthy
        operation.) Table will be reopened on a subsequent call to parmtable() below."""
        if self._pt:
            self._pt = None

    def parmtable(self, write=False):
        """Returns FastParmTable object, reopening it if needed. Use write=True to reopen for writing."""
        if not self.filename:
            return None
        if self._pt and write and not self._pt_write:
            self._pt = None
        if not self._pt:
            self._pt = FastParmTable(self.filename, write)
            self._pt_write = write
        return self._pt

    def funklet_names(self):
        """Returns list of funklet names"""
        return self._funklet_names

    def funklet_name_components(self):
        """Returns list of sets of funklet name components"""
        return self._name_components

    def envelope_domain(self):
        """Returns domain object which envelops all of the parmtable"""
        kw = {}
        for iaxis, stats in enumerate(self._axis_stats):
            if not stats.empty():
                kw[mequtils.get_axis_id(iaxis)] = stats.minmax
        return meq.gen_domain(**kw)

    def envelope_cells(self, **num_cells):
        """Returns cells object which envelops all of the parmtable, and is regularly spaced. The number of points
        along each axis is equal to the number of subdomains along that axis, but the cells do not necessarily follow
        the structure of the subdomain if the subdomains are overlapping, spaced out, or irregular.
        """
        dom = self.envelope_domain()
        kw = {}
        for iaxis, stats in enumerate(self._axis_stats):
            if not stats.empty():
                kwname = 'num_' + str(mequtils.get_axis_id(iaxis)).lower()
                kw[kwname] = num_cells.get(kwname, len(stats.cells))
        return meq.gen_cells(dom, **kw)

    def subdomain_cells(self):
        """Returns cells object which envelops all of the parmtable, and contains a cells matching every subdomain."""
        dom = self.envelope_domain()
        kw = {}
        cells = meq.gen_cells(dom)
        for iaxis, stats in enumerate(self._axis_stats):
            if not stats.empty():
                grid = list(stats.cells.keys())
                grid.sort()
                meq.add_cells_axis(cells, mequtils.get_axis_id(iaxis), grid=grid, cell_size=[
                    stats.cells[x0] for x0 in grid])
        return cells

    def axis_stats(self, iaxis):
        """Returns _AxisStats object for the specified parmtable."""
        return self._axis_stats[iaxis]

    def _make_axis_index(self):
        """Builds up various indices based on content of the parmtable"""
        # check if cache is up-to-date
        cachepath = os.path.join(self.filename, 'ParmTab.cache')
        funkpath = os.path.join(self.filename, 'funklets')
        self.mtime = os.path.getmtime(funkpath) if os.path.exists(funkpath) else time.time()
        try:
            has_cache = os.path.getmtime(cachepath) >= self.mtime
            if not has_cache:
                dprintf(2, "cache is out of date, will regenerate\n")
        except:
            dprintf(0, "%s: os.path.getmtime() throws exception, assuming cache is out of date\n", self.filename)
            has_cache = False
        # try to load the cache if so
        t0 = time.time()
        if has_cache:
            try:
                dprintf(2, "loading index cache\n")
                self._funklet_names, self._domain_list, self._axis_stats, self._name_components, \
                self._domain_fullset, self._domain_cell_index, self._domain_reverse_index \
                    = pickle.load(open(cachepath, 'rb'))
                dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
                t0 = time.time()
                return
            except:
                if verbosity.get_verbose() > 0:
                    traceback.print_exc()
                dprintf(0, "%s: error reading cached stats, regenerating\n", self.filename)
                has_cache = False
        # no cache, so regenerate everything
        if not has_cache:
            self._axis_stats = [_AxisStats(mequtils.get_axis_id(i)) for i in range(mequtils.max_axis)]
            pt = self.parmtable()
            dprintf(2, "loading domain list\n")
            self._domain_list = pt.domain_list()
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()
            dprintf(2, "collecting axis stats\n")
            self._axes = {}
            for domain in self._domain_list:
                for axis, rng in list(domain.items()):
                    if str(axis) != 'axis_map':
                        self._axis_stats[mequtils.get_axis_number(axis)].add_cell(*rng)
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()
            dprintf(2, "finalizing axis stats\n")
            self._domain_fullset = [None] * mequtils.max_axis
            for iaxis, stats in enumerate(self._axis_stats):
                stats.update()
                if not stats.empty():
                    self._domain_fullset[iaxis] = list(range(len(stats.cells)))
                    dprintf(2, "axis %s: %d unique cells from %g to %g\n", stats.name, len(stats.cells), *stats.minmax)
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()
            dprintf(2, "making subdomain indices\n")
            # now make a subdomain index
            self._domain_cell_index = [0] * len(self._domain_list)
            self._domain_reverse_index = {}
            for idom, domain in enumerate(self._domain_list):
                index = [None] * mequtils.max_axis
                for axis, rng in list(domain.items()):
                    if str(axis) != 'axis_map':
                        iaxis = mequtils.get_axis_number(axis)
                        index[iaxis] = self._axis_stats[iaxis].lookup_cell(*rng)
                # insert into domain_cells_index and domain_reverse_index
                index = tuple(index)
                self._domain_cell_index[idom] = index
                self._domain_reverse_index[index] = idom
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()

            dprintf(2, "loading funklet name list\n")
            self._funklet_names = list(pt.name_list())
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()
            dprintf(2, "computing funklet indices\n")
            self._name_components = {}
            for name in self._funklet_names:
                for i, token in enumerate(name.split(':')):
                    self._name_components.setdefault(i, set()).add(token)
            self._name_components = [self._name_components[i] for i in range(len(self._name_components))]
            for i, values in enumerate(self._name_components):
                dprintf(2, "component %d: %s\n", i, ' '.join(values))
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()

            dprintf(2, "writing cache\n")
            try:
                pickle.dump((
                    self._funklet_names, self._domain_list, self._axis_stats, self._name_components, \
                    self._domain_fullset, self._domain_cell_index, self._domain_reverse_index \
                    ), open(cachepath, 'wb')
                )
            except:
                if verbosity.get_verbose() is not None and verbosity.get_verbose() > 0:
                    traceback.print_exc()
                dprintf(0, "%s: error writing stats to cache, will probably regenerate next time\n", self.filename)
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()

    def resolve_output_table(self, outtab, new=False):
        """Helper function. Resolves outtab to a ParmTab object as follows:
        * if it is a ParmTab, returns it directly
        * if it is a filename, creates ParmTab for that table
        * If it is None, creates a ParmTab based on the current table.
        """
        if isinstance(outtab, ParmTab):
            return outtab
        elif isinstance(outtab, str):
            return ParmTab(outtab, write=True, new=new)
        else:
            name, ext = os.path.splitext(self.filename)
            return ParmTab(name + "_out" + ext, write=True, new=new)

    def make_slicing(self, slicing):
        if not slicing:
            return DomainSlicing([], self._domain_fullset)
        elif isinstance(slicing, DomainSlicing):
            return slicing
        elif isinstance(slicing, (str, int)):
            slicing = [slicing]
        return DomainSlicing(slicing, self._domain_fullset)

    def funkset(self, name):
        return FunkSet(self, name)

    def apply(self, op_func, slicing, outtab=None, remove=False, newtab=False):
        """For each funklet in our table, takes all funklets along the designated slicing axis
        (i.e. for each slice along the non-listed axes), creates a FunkSlice, and calls
        op_func(slice).
        The return value of func() should be a funklet list, which is written to the output table,
        or None to ignore.
        if 'outtab' is None, a new output table is created. Otherwise set outtab to a filename, or a
        ParmTab, or a FastParmTable.
        If 'remove' is True, input funklets will be removed if an output funklet is returned.
        """
        t0 = time.time()
        nn = len(self.funklet_names())
        self._start_progress("applying operation '%s'" % op_func.__name__, nn)
        try:
            outtab = self.resolve_output_table(outtab, newtab)
            dprintf(1, "using output table %s\n", outtab.filename)
            pt = self.parmtable()
            # loop over funklets and slices
            dprintf(3, "input slicing is %s\n", slicing)
            slicing = self.make_slicing(slicing)
            dprintf(2, "%d slices will be iterated over\n", len(slicing))
            for iname, name in enumerate(sort_qualified_names(self.funklet_names())):
                self._report_progress(iname)
                self.funkset(name).apply(op_func, slicing, outtab=outtab, remove=remove)
        finally:
            dprintf(2, "elapsed time: %f seconds\n", time.time() - t0)
            t0 = time.time()
            self._end_progress(False)
            self.close()


def open(*args, **kw):
    """Opens a ParmTab. Arguments are passed to ParmTab constructor:
  
    """ + ParmTab.__init__.__doc__
    return ParmTab(*args, **kw)


if __name__ == '__main__':
    import sys
    from . import FunkOps

    if len(sys.argv) < 2:
        print("Pass in an fmep table to read its stats")
    else:
        verbose(3)
        pt = ParmTab(sys.argv[1])
        if '-average' in sys.argv:
            pt.apply(FunkOps.average, "time", newtab=True)
        if '-interpol' in sys.argv:
            pt.apply(FunkOps.linear_interpol, "time", newtab=True)
        if '-rank0' in sys.argv:
            pt.apply(FunkOps.force_rank0, ["time", "freq"], newtab=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time
import math
import Owlcat

ROWCHUNK = 50000

time_start = time.time()

import pyrap.tables
from casacore.tables import table


def timestamp(format="%H:%M:%S:"):
    return time.strftime(format, time.gmtime(time.time() - time_start))


def progress(message, newline=True):
    sys.stdout.write("%s %-60s%s" % (timestamp(), message, '\n' if newline else '\r'))
    sys.stdout.flush()


def call_system(command, args):
    rc = os.system("%s %s" % (command, args))
    if rc:
        stat = rc >> 8
        sig = rc & 0xFF
        if sig:
            print(("%s process killed by signal %d" % (command, sig)))
        else:
            print(("%s call failed with exit status", (command, stat)))
            if stat == 127:
                print(("Please check your MeqTrees installation for the %s utility!" % command))
    return rc


if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser

    parser = OptionParser(usage="""%prog: [options] MS""",
                          description= \
                              "Sets up tiled storage managers for the DATA, FLAG and BITFLAG columns of an MS, " +
                              "and sets a fixed shape on the columns. This can speed up access to some MSs, e.g. " +
                              "such as those produced by the uvfits2ms converter. "
                          )
    parser.add_option("-f", "--force", action="store_true",
                      help="proceed non-interactively")

    (options, msnames) = parser.parse_args()

    if len(msnames) != 1:
        parser.error("MS name not supplied.")

    if not Owlcat.find_exec("addtiledmscol"):
        print("Can't find the 'addtiledmscol' utility, which is required by this script.")
        print("You may need to install or build the MeqTrees package.")
        sys.exit(1)

    msname = msnames[0]
    ms = table(msname, readonly=False)

    if not options.force:
        print("""
  This script will reinitialize the DATA, FLAG and BITFLAG columns
  of the MS using a tiled storage manager and a fixed shape. It will attempt
  to preserve the data in these columns. The MODEL_DATA, CORRECTED_DATA and
  IMAGING_WEIGHT columns will be reset and their contents discarded. If this
  operation is interrupted, the MS may be left in a corrupted state, so
  please make sure you have a backup of the MS.
  """)
        if input("Proceed (y/n)? ").strip().upper()[0] != "Y":
            print("Aborted by user.")
            sys.exit(1)

    SAVECOLS = ["DATA", "FLAG", "BITFLAG"]
    KILLCOLS = SAVECOLS + ["MODEL_DATA", "CORRECTED_DATA", "IMAGING_WEIGHT"]

    # read in DATA and FLAG cols
    savedcols = {}
    for colname in "DATA", "FLAG", "BITFLAG":
        if colname in ms.colnames():
            progress("Reading %s column" % colname)
            savedcols[colname] = ms.getcol(colname)
        elif colname is "DATA":
            print("Error: this MS does not appear to have a DATA column")
            sys.exit(1)

    datacol = savedcols['DATA']
    progress("DATA column shape is %s" % (datacol.shape,))
    flagcol = savedcols.get('FLAG')
    if flagcol is not None and datacol.shape != flagcol.shape:
        print(("Error: FLAG shape (%s) does not match DATA shape" % (flagcol.shape,)))

    # kill columns
    cols = [col for col in KILLCOLS if col in ms.colnames()]
    if cols:
        progress("Removing columns %s" % ", ".join(cols))
        ms.removecols(cols)

    # now reinsert DATA and FLAG columns
    ms.close()
    ncorr = datacol.shape[-1]
    nfreq = datacol.shape[-2]
    tilerow = 512
    for colname, coltype in ("DATA", "complex"), ("FLAG", "bool"), ("BITFLAG", "int"):
        if colname in savedcols:
            progress("Inserting tiled %s column" % colname)
            if not call_system("addtiledmscol", "%s %s %s %d %d %d %d %d 4" %
                                                (msname, colname, coltype, ncorr, nfreq, min(ncorr, 4), min(nfreq, 8),
                                                 tilerow)):
                sys.exit(1)

    # reopen table
    ms = table(msname, readonly=False)
    for colname in SAVECOLS:
        col = savedcols.get(colname)
        if col is not None:
            progress("Rewriting %s column" % colname)
            ms.putcol(colname, col)

    progress("Closing MS")
    ms.close()
    progress("Adding imaging columns")
    pyrap.tables.addImagingColumns(msname)
    progress("Done")



# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import print_function
import sys
import re
import os.path
import numpy
from astropy.io import fits as pyfits
import scipy.ndimage.measurements
import math
#from astLib.astWCS import WCS
from astropy.wcs import WCS
import glob

SANITIZE_DEFAULT = 12345e-7689


def stack_planes(fitslist, outname='combined.fits', axis=0, ctype=None, keep_old=False, fits=False):
    """ Stack a list of fits files along a given axiis.
       
       fitslist: list of fits file to combine
       outname: output file name
       axis: axis along which to combine the files
       fits: If True will axis FITS ordering axes
       ctype: Axis label in the fits header (if given, axis will be ignored)
       keep_old: Keep component files after combining?
    """

    hdu = pyfits.open(fitslist[0])[0]
    hdr = hdu.header
    naxis = hdr['NAXIS']

    # find axis via CTYPE key
    if ctype is not None:
        for i in range(1, naxis + 1):
            if hdr['CTYPE%d' % i].lower().startswith(ctype.lower()):
                axis = naxis - i  # fits to numpy convention
    elif fits:
        axis = naxis - axis

    fits_ind = abs(axis - naxis)
    crval = hdr['CRVAL%d' % fits_ind]

    imslice = [slice(None)] * naxis
    _sorted = sorted([pyfits.open(fits) for fits in fitslist],
                     key=lambda a: a[0].header['CRVAL%d' % (naxis - axis)])

    # define structure of new FITS file
    nn = [hd[0].header['NAXIS%d' % (naxis - axis)] for hd in _sorted]
    shape = list(hdu.data.shape)
    shape[axis] = sum(nn)
    data = numpy.zeros(shape, dtype=float)

    for i, hdu0 in enumerate(_sorted):
        h = hdu0[0].header
        d = hdu0[0].data
        imslice[axis] = list(range(sum(nn[:i]), sum(nn[:i + 1])))
        data[imslice] = d
        if crval > h['CRVAL%d' % fits_ind]:
            crval = h['CRVAL%d' % fits_ind]
        try:
            hdr['BMAJ%d' % (i + 1)] = h['BMAJ']
            hdr['BMIN%d' % (i + 1)] = h['BMIN']
            hdr['BPA%d' % (i + 1)] = h['BPA']
        except KeyError:
            pass

    # update header
    hdr['CRVAL%d' % fits_ind] = crval
    hdr['CRPIX%d' % fits_ind] = 1

    pyfits.writeto(outname, data.astype(numpy.float32), hdr, overwrite=True)
    print(("Successfully stacked images. Output image is %s" % outname))

    # remove old files
    if not keep_old:
        for fits in fitslist:
            os.system('rm -f %s' % fits)


def unstack_planes(fitsname, each_chunk, axis=None, ctype=None, prefix=None, fits=False, keep_old=True):
    """ 
        Unstack FITS image along a given axis into multiple 
        images each having each_chunk planes along that axis 
    """

    prefix = prefix or fitsname[:-5]  # take everthing but .FITS/.fits
    hdu = pyfits.open(fitsname)
    hdr = hdu[0].header
    data = hdu[0].data.copy()
    naxis = hdr["NAXIS"]

    if axis is None and ctype is None:
        raise SystemExit('Please specify either axis or ctype')
    # find axis via CTYPE key

    if ctype:
        for i in range(1, naxis + 1):
            if hdr['CTYPE%d' % i].lower().startswith(ctype.lower()):
                axis = naxis - i  # fits to numpy indexing
    elif fits:
        axis = naxis - axis

    crval = hdr['CRVAL%d' % (naxis - axis)]
    cdelt = hdr['CDELT%d' % (naxis - axis)]
    crpix = hdr['CRPIX%d' % (naxis - axis)]
    # shift crval to crpix=1
    crval = crval - (crpix - 1) * cdelt

    nstacks = hdr['NAXIS%d' % (naxis - axis)]
    nchunks = nstacks // each_chunk
    print(("The FITS file %s has %s stacks along this axis. Breaking it up to %d images" % (fitsname, nstacks, nchunks)))

    outfiles = []
    for i in range(0, nchunks):
        _slice = [slice(None)] * naxis
        _slice[axis] = list(range(i * each_chunk, (i + 1) * each_chunk if i + 1 != nchunks else nstacks))
        hdu[0].data = data[_slice].astype(numpy.float32)
        hdu[0].header['CRVAL%d' % (naxis - axis)] = crval + i * cdelt * each_chunk
        hdu[0].header['CRPIX%d' % (naxis - axis)] = 1
        outfile = '%s-%04d.fits' % (prefix, i)
        outfiles.append(outfile)
        print(("Making chunk %d : %s. File is %s" % (i, repr(_slice[axis]), outfile)))
        hdu.writeto(outfile, overwrite=True)
    hdu.close()

    if keep_old == False:
        os.system('rm -rf %s' % fitsname)

    return outfiles


def reorder(fitsname, order=[], outfile=None):
    """ 
    Re-order FITS image axes. 

    Example:
        If your FITS files has axes (NAXIS{1,2,3,4})-> RA, DEC, STOKES, FREQ
        You can re-order it so that it has RA, DEC, FREQ, STOKES
        by running: reorder(fitsname, order=[1,2,4,3], "my_reodered_image.fits")
    """

    # Get to know input FITS image
    hdu = pyfits.open(fitsname)
    hdr0 = hdu[0].header
    ndim = hdr0["NAXIS"]
    order0 = list(range(1, ndim + 1))

    def fits2py(a):
        if isinstance(a, (list, tuple)):
            return list(ndim - numpy.array(a))
        else:
            return ndim - a

    if order == order0:
        print("The FITS image already has the order you requested")
        return
    elif len(order) != ndim:
        raise ValueError("The dimensions of the FITS image do not match your request.")

    # Ok, Lock and Load
    data = hdu[0].data.astype(numpy.float32)
    # First re-order data
    hdu[0].data = numpy.transpose(data, list((ndim - numpy.array(order))[list(range(ndim - 1, -1, -1))]))

    hdr = hdr0.copy()
    mendatory = "CTYPE CRVAL CDELT CRPIX".split()
    optional = "CUNIT CROTA".split()

    for key in mendatory + optional:
        for old, new in zip(order0, order):
            if old == new:
                continue
            try:
                val = hdr0["%s%d" % (key, old)]
                comment = hdr0.comments["%s%d" % (key, old)]
                idx = hdr.index("%s%d" % (key, old))
                # Seems like can't replace non empy val with an empty string. So delete first

                del hdr["%s%d" % (key, new)]
                hdr.insert(idx, ("%s%d" % (key, new), val, comment))

            except KeyError:
                if key in mendatory:
                    raise KeyError("ARBORTNG: FITS file doesn't have the '%s' key in the header" % key)
                else:
                    pass

    hdu[0].header = hdr
    outfile = outfile or "reodered_" + fitsname
    print(("Successfully re-ordered axes in FITS image. Output image is at %s" % outfile))
    hdu.writeto(outfile, overwrite=True)


def main():
    # setup some standard command-line option parsing
    #
    from optparse import OptionParser
    parser = OptionParser(usage="""%prog: [options] <image names...>""")
    parser.add_option("-o", "--output", dest="output", type="string",
                      help="name of output FITS file")
    parser.add_option("-r", "--replace", action="store_true",
                      help="replace (first) input file by output. Implies '--force'.")
    parser.add_option("-f", "--force", dest="force", action="store_true",
                      help="overwrite output file even if it exists")
    parser.add_option("-S", "--sanitize", type="float", metavar="VALUE",
                      help="sanitize FITS files by replacing NANs and INFs with VALUE")
    parser.add_option("-Z", "--zero-to-nan", action='store_true',
                      help="Replace zeros with NaNs")
    parser.add_option("-N", "--nonneg", action="store_true",
                      help="replace negative values by 0")
    parser.add_option("-m", "--mean", dest="mean", action="store_true",
                      help="take mean of input images")
    parser.add_option("-d", "--diff", dest="diff", action="store_true",
                      help="take difference of 2 input images")
    parser.add_option("--div", dest="div", action="store_true",
                      help="take ratio of 2 input images (e.g. as in primary beam correction)")
    parser.add_option("--div-min", type="float",
                      help="min value to use for second image, when taking a ratio")
    parser.add_option("--div-min-mask", type="float", metavar="MIN", default=0,
                      help="mask value to apply to output where second image < MIN")
    parser.add_option("--sum", dest="sum", action="store_true",
                      help="sum of input images")
    parser.add_option("--prod", dest="prod", action="store_true",
                      help="product of input images")
    parser.add_option("-t", "--transfer", action="store_true",
                      help="transfer data from image 2 into image 1, preserving the FITS header of image 1")
    parser.add_option("-z", "--zoom", dest="zoom", type="int", metavar="NPIX",
                      help="zoom into central region of NPIX x NPIX size")
    parser.add_option("-p", "--paste", action="store_true",
                      help="paste image(s) into first image, using the nearest WCS pixel positions. See also --empty-canvas")
    parser.add_option("--empty-canvas", action="store_true",
                      help="null first image, then paste image(s) into it")
    parser.add_option("--null-region", type="str", metavar="NXxNY", help="null central region of specified size")
    parser.add_option("-M", "--set-max", metavar="MAX", type="float",
                      help="set values over MAX to MAX")
    parser.add_option("-R", "--rescale", dest="rescale", type="float",
                      help="rescale image values")
    parser.add_option("-E", "--edit-header", metavar="KEY=VALUE", type="string", action="append",
                      help="replace header KEY with VALUE. Use KEY=VALUE for floats and KEY='VALUE' for strings.")
    parser.add_option("-D", "--delete-header", metavar="KEY", type="string", action="append",
                      help="Remove KEY from header")
    parser.add_option("--stack", metavar="outfile:axis",
                      help="Stack a list of FITS images along a given axis. This axis may given as an integer"
                           "(as it appears in the NAXIS keyword), or as a string (as it appears in the CTYPE keyword)")
    parser.add_option("--reorder",
                      help="Required order. List of comma seperated indeces")
    parser.add_option("--add-axis", metavar="CTYPE:CRVAL:CRPIX:CDELT[:CUNIT:CROTA]", action="append", default=[],
                      help="Add axis to a FITS image. The AXIS will be described by CTYPE:CRVAL:CRPIX:CDELT[:CUNIT:CROTA]. "
                           "The keywords in brackets are optinal, while those not in brackets are mendatory. "
                           "This axis will be the last dimension. Maybe specified more than once.")
    parser.add_option("--unstack", metavar="prefix:axis:each_chunk",
                      help="Unstack a FITS image into smaller chunks each having [each_chunk] planes along a given axis. "
                           "This axis may given as an integer (as it appears in the NAXIS keyword), or as a string "
                           "(as it appears in the CTYPE keyword)")
    parser.add_option("--delete-files", action="store_true",
                      help="Delete original file(s) after stacking/unstacking using --stack/--unstack")
    parser.add_option("-H", "--header", action="store_true", help="print header(s) of input image(s)")
    parser.add_option("-s", "--stats", action="store_true",
                      help="print stats on images and exit. No output images will be written.")
    parser.add_option("-F", "--file_pattern",
                      help="Speicfy input images via a pattern string, e.g, \"prefix*June2016.fits\". NB: The qouatation marks are important.")

    parser.set_defaults(output="", mean=False, zoom=0, rescale=1, edit_header=[], delete_header=[])

    (options, imagenames) = parser.parse_args()

    if options.file_pattern:
        imagenames = glob.glob(options.file_pattern)

    if not imagenames:
        parser.error("No images specified. Use '-h' for help.")

    # print "%d input image(s): %s"%(len(imagenames),", ".join(imagenames))
    images = [pyfits.open(img) for img in imagenames]
    updated = False

    # Stack FITS images
    if options.stack:
        if len(imagenames) < 1:
            parser.error("Need more than one image to stack")
        stack_args = options.stack.split(":")
        if len(stack_args) != 2:
            parser.error("Two --stack options are required. See ./fitstool.py -h")

        outfile, axis = stack_args

        _string = True
        try:
            axis = int(axis)
            _string = False
        except ValueError:
            _string = True

        stack_planes(imagenames, ctype=axis if _string else None, keep_old=not options.delete_files,
                     axis=None if _string else axis, outname=outfile, fits=True)

    # Unstack FITS image
    if options.unstack:
        image = imagenames[0]
        if len(imagenames) < 1:
            parser.error("Need more than one image to stack")
        unstack_args = options.unstack.split(":")
        if len(unstack_args) != 3:
            parser.error("Two --unstack options are required. See ./fitstool.py -h")

        prefix, axis, each_chunk = unstack_args

        _string = True
        try:
            axis = int(axis)
            _string = False
        except ValueError:
            _string = True

        each_chunk = int(each_chunk)

        unstack_planes(image, each_chunk, ctype=axis if _string else None,
                       axis=None if _string else axis, prefix=prefix, fits=True,
                       keep_old=not options.delete_files)
        sys.exit(0)

    if options.add_axis:
        for axis in options.add_axis:
            hdu = pyfits.open(imagenames[0])
            hdr = hdu[0].header
            ndim = hdr["NAXIS"]
            hdu[0].data = hdu[0].data[numpy.newaxis, ...]
            _mendatory = "CTYPE CRVAL CDELT CRPIX".split()
            _optional = "CUNIT CROTA".split()
            L = len(_mendatory + _optional)
            values = axis.split(":")
            Lv = len(values)

            if len(_mendatory) > len(values):
                parser.error("Something with the way specified --add-axis. See %prog -h for help")

            for i, value in enumerate(values):
                try:
                    value = float(value)
                except ValueError:
                    if isinstance(value, str):
                        value = value.upper()

                hdu[0].header["%s%d" % ((_mendatory + _optional)[i], ndim + 1)] = value

            hdu.writeto(imagenames[0], overwrite=True)
            print(("Successfully added axis %s to %s" % (values[0], imagenames[0])))

    if options.reorder:
        for image in imagenames:
            order = list(map(int, options.reorder.split(",")))
            reorder(image, order=order, outfile=image)

    if options.header:
        for filename, img in zip(imagenames, images):
            img.verify('silentfix')
            if len(imagenames) > 1:
                print("======== FITS header for", filename)
            for hdrline in img[0].header.cards:
                print(hdrline)

    if options.replace:
        if options.output:
            parser.error("Cannot combine -r/--replace with -o/--output")
        outname = options.output or imagenames[0]
        options.force = True
        autoname = False
    else:
        outname = options.output
        autoname = not outname
        if autoname:
            outname = re.split('[_]', imagenames[0], 1)[-1]

    for keyval in options.edit_header:
        key, val = keyval.split("=")
        q = ''
        if val[0] == "'" and val[-1] == "'":
            images[0][0].header[key] = val[1:-1:]
            q = '"'
        elif val[-1] == 'd' or key.startswith('NAXIS'):
            images[0][0].header[key] = int(val[:-1] if val[-1] == 'd' else val)
        else:
            try:
                images[0][0].header[key] = float(val)
            except:
                images[0][0].header[key] = val
                q = '"'
        print("Setting header %s=%s%s%s" % (key, q, val, q))
        updated = True

    for key in options.delete_header:
        try:
            del images[0][0].header[key]
        except KeyError:
            raise "Key '%s' not found in FITS header" % key
        print("Deleting key '%s' header" % key)
        updated = True

    if options.sanitize is not None:
        print("Sanitizing: replacing INF/NAN with", options.sanitize)
        for img in images:
            d = img[0].data
            d[numpy.isnan(d) + numpy.isinf(d)] = options.sanitize
        # if using stats, do not generate output
        if not options.stats:
            updated = True

    if options.zero_to_nan:
        print("Replacing zeros with NaNs")
        for img in images:
            d = img[0].data
            d[d == 0] = numpy.nan
        # if using stats, do not generate output
        if not options.stats:
            updated = True

    if options.nonneg:
        print("Replacing negative value by 0")
        for img, name in zip(images, imagenames)[:1]:
            d = img[0].data
            wh = d < 0
            d[wh] = 0
            print("Image %s: replaced %d points" % (name, wh.sum()))
        updated = True

    def compute_in_out_slice(N, N0, R, R0):
        """
        given an input axis of size N, and an output axis of size N0, and reference pixels of I and I0
        respectively, computes two slice objects such that
                A0[slice_out] = A1[slice_in]
        would do the correct assignment (with I mapping to I0, and the overlapping regions transferred)
        """
        i, j = 0, N     # input slice
        i0 = R0 - R
        j0 = i0 + N     # output slice
        if i0 < 0:
            i = -i0
            i0 = 0
        if j0 > N0:
            j = N - (j0 - N0)
            j0 = N0
        if i >= j or i0 >= j0:
            return None, None
        return slice(i0, j0), slice(i, j)

    if options.transfer:
        if len(images) != 2:
            parser.error("The --transfer option requires exactly two input images.")
        if autoname:
            outname = "xfer_" + outname
        print("Transferring %s into coordinate system of %s" % (imagenames[1], imagenames[0]))
        images[0][0].data = images[1][0].data
        updated = True
    elif options.diff:
        if len(images) != 2:
            parser.error("The --diff option requires exactly two input images.")
        if autoname:
            outname = "diff_" + outname
        print("Computing difference")
        data = images[0][0].data
        data -= images[1][0].data
        updated = True
    elif options.sum:
        if autoname:
            outname = "sum_" + outname
        print("Computing sum")
        data = images[0][0].data
        for d in images[1:]:
            data += d[0].data
        updated = True
    elif options.div:
        if len(images) != 2:
            parser.error("The --div option requires exactly two input images.")
        if autoname:
            outname = "div_" + outname
        print("Computing ratio")
        a, b = images[0][0].data, images[1][0].data
        if options.div_min is not None:
            b[b<=options.div_min] = options.div_min
        a /= b
        if options.div_min_mask is not None:
            a[b<=options.div_min] = options.div_min_mask
        updated = True
    elif options.prod:
        if autoname:
            outname = "prod_" + outname
        print("Computing product")
        data = images[0][0].data
        for d in images[1:]:
            data *= d[0].data
        updated = True
    elif options.mean:
        if autoname:
            outname = "mean%d_" % len(images) + outname
        print("Computing mean")
        data = images[0][0].data
        for img in images[1:]:
            data += img[0].data
        data /= len(images)
        images = [images[0]]
        updated = True
    elif options.paste:
        if autoname:
            outname = "paste%d_" % len(images) + outname
        print("Pasting images into canvas given by {}".format(imagenames[0]))
        print("WARNING: the reference pixel of the pasted image will be aligned with the WCS of the canvas image,")
        print("         but no WCS reprojection will be attempted. Results may be grossly incorrect if your WCSs")
        print("         are sufficiently different. Use at own risk.")
        # get original image size and WCS
        data0 = images[0][0].data
        ny0, nx0 = data0.shape[-2:]
        hdr0 = images[0][0].header
        wcs0 = WCS(hdr0, mode="pyfits")
        if options.empty_canvas:
            print("  canvas image will be cleared initially")
            data0[:] = 0

        for image, imagename in zip(images[1:], imagenames[1:]):
            hdr = image[0].header
            data = image[0].data
            ny, nx = data.shape[-2:]
            rx, ry = int(hdr["CRPIX1"]-1), int(hdr["CRPIX2"]-1)
            # find coordinate of this image's reference pixel in the canvas WCS
            rx0, ry0 = wcs0.wcs2pix(hdr["CRVAL1"], hdr["CRVAL2"])  # get WCS of center pixel
            print("image {} rpix {},{} is at {},{} in canvas image".format(imagename, rx, ry, rx0, ry0))
            rx0 = int(round(rx0))
            ry0 = int(round(ry0))
            xout, xin = compute_in_out_slice(nx, nx0, rx, rx0)
            yout, yin = compute_in_out_slice(ny, ny0, ry, ry0)
            if xout is not None and yout is not None:
                print("  canvas X {} will be assigned from {}".format(xout, xin))
                print("  canvas Y {} will be assigned from {}".format(yout, yin))
                data0[..., yout, xout] += data[..., yin, xin]
                updated = True
            else:
                print("  image {} footprint is outside of canvas".format(imagename))

    if options.null_region:
        if len(images) > 1:
            print("Too many input images specified for this operation, at most 1 expected")
            sys.exit(2)
        data = images[0][0].data
        nx = data.shape[-1]
        ny = data.shape[-2]
        xc, yc = nx//2, ny//2
        rx, ry = map(int,options.null_region.split("x", 1))
        rx, ry = min(rx, nx), min(ry, ny)
        x0, y0 = xc - rx//2, yc - ry//2
        data[..., y0:y0+ry, x0:x0+rx] = 0
        print("Setting {}:{} {}:{} to zero".format(x0, x0+rx, y0, y0+ry))
        updated = True

    if options.zoom:
        z = options.zoom
        if autoname:
            outname = "zoom%d_" % z + outname
        if len(images) > 1:
            print("Too many input images specified for this operation, at most 1 expected")
            sys.exit(2)
        data = images[0][0].data
        nx = data.shape[-1]
        ny = data.shape[-2]
        # make output array of shape z x z
        zdata_shape = list(data.shape)
        zdata_shape[-1] = zdata_shape[-2] = z
        zdata = numpy.zeros(zdata_shape, dtype=data.dtype)
        # make input/output slices
        rx, ry = nx//2, ny//2
        rx0 = ry0 = z//2
        xout, xin = compute_in_out_slice(nx, z, rx, rx0)
        yout, yin = compute_in_out_slice(ny, z, ry, ry0)

        zdata[..., yout, xout] = data[..., yin, xin]

        # update header
        hdr = images[0][0].header
        # hdr1 = hdr.copy()
        # hdr1["NAXIS"] = 2
        # # newer astropy don't like our radio axes 3 and 4 :((
        # for nax in "3", "4":
        #     for key in "NAXIS", "CTYPE", "CRVAL", "CRPIX", "CDELT", "CROTA":
        #         hdr1.remove(key + nax, ignore_missing=True)
        wcs = WCS(hdr) #, mode="pyfits")
        pixcoord = numpy.zeros((1, hdr["NAXIS"]), float)
        pixcoord[0, 0] = nx//2
        pixcoord[0, 1] = ny//2
        world = wcs.wcs_pix2world(pixcoord, 0)  # get WCS of center pixel
        hdr["CRVAL1"] = world[0,0]
        hdr["CRVAL2"] = world[0,1]
        hdr["CRPIX1"] = rx0 + 1
        hdr["CRPIX2"] = ry0 + 1

        print("Making zoomed image of shape", "x".join(map(str, zdata.shape)))

        images = [pyfits.PrimaryHDU(zdata, hdr)]
        updated = True

    if options.set_max is not None:
        print("Enforcing maximum value of {}".format(options.set_max))
        images[0][0].data[images[0][0].data > options.set_max] = options.set_max
        updated = True

    if options.rescale != 1:
        if autoname and not updated:
            outname = "rescale_" + outname
        if len(images) > 1:
            "Too many input images specified for this operation, at most 1 expected"
            sys.exit(2)
        print("Applying scaling factor of %f to image values" % options.rescale)
        images[0][0].data *= options.rescale
        updated = True

    if updated:
        imagenames[0] = outname

    if options.stats:
        for ff, filename in zip(images, imagenames):
            data = ff[0].data
            min1, max1, dum1, dum2 = scipy.ndimage.measurements.extrema(data)
            sum = data.sum()
            mean = sum / data.size
            std = math.sqrt(((data - mean) ** 2).mean())
            print("%s: min %g, max %g, sum %g, np %d, mean %g, std %g" % (filename, min1, max1, sum, data.size, mean, std))
        sys.exit(0)

    if updated:
        print("Writing output image", outname)
        if os.path.exists(outname) and not options.force:
            print("Output image exists, rerun with the -f switch to overwrite.")
            sys.exit(1)
        images[0].writeto(outname, overwrite=True)
    elif not (options.header or options.stack or options.add_axis or options.reorder):
        print("No operations specified. Use --help for help.")



from argparse import ArgumentParser
from Owlcat import PlotAntennas as pants


def get_arguments():
    """Command line arguments"""

    parser = ArgumentParser(usage="""%(prog)s: [options] ms""",
                            description="Plot antenna positions")

    parser.add_argument("ms_name", metavar='',
                        help="MS / Antenna table name",
                        default=None)
    parser.add_argument("-c", "--colour", dest="cmap",  metavar='', type=str,
                        help="""Show antennas used in selected MS in this
                        colour. Can be specified as colour name or
                        hex / rgb / rgba. If specified as hex value or rgb,
                        use quotes. e.g '#1ECBE1' or 'rgb(255, 0, 0)'. """,
                        default=None)
    parser.add_argument("--cofa", dest="cofa", metavar='',
                        type=str,
                        help="""Coordinates for the centre of array in the
                        MS. If supplied, it should be specified as a comma
                        separated string in the order lon,lat,elevation. 
                        Lon and Lat values should be in degrees, while
                        elevation should be in metres. If no elevation is
                        available, set the value of 0.""",
                        default=None)
    parser.add_argument("-o", "--output", dest="o_file",  metavar='',
                        type=str,
                        help="""Name to give output html file including the
                        file extension. e.g. 'test.html'. Default is
                        msName.html""", default=None)
    parser.add_argument("-uf", "--underlay-file", dest="u_file", metavar='',
                        type=str,
                        help="""JSON File containing a telescope's antennas'
                     coordinates. If supplied, this script will plot the
                     locations of all the antennas in the file, in addition
                     to the antennas that were used in the specified MS. An
                     example of such a file can be found in the data
                     directory of this package.
                     """, default=None)

    return parser


if __name__ == "__main__":

    parser = get_arguments()
    args = parser.parse_args()
    pants.main(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup
    from Owlcat.Plotting import LSTElevationPlot
    import casacore.tables

    parser = OptionParser(usage="""%prog: [plots & options] MS""",
                          description="""Plots elevation tracks from an MS.""")

    parser.add_option("-l", "--list", action="store_true",
                      help="lists fields found in MS, then exits")

    plotgroup = OptionGroup(parser, "Plotting options")
    outputgroup = OptionGroup(parser, "Output options")
    LSTElevationPlot.init_options(plotgroup, outputgroup)

    outputgroup.add_option("-o", "--output-name", type=str, help="Output filename", default="lst-elev.png")
    outputgroup.add_option("-d", "--display", action="store_true", help="Display plot on screen")

    parser.add_option_group(plotgroup)
    parser.add_option_group(outputgroup)

    (options, args) = parser.parse_args()

    if not args:
        parser.error("No MS specified")

    if options.display:
        options.output_name = None

    skyplot = LSTElevationPlot(options, output_type='x11' if options.display else 'png')

    for msname in args:
        field_time, field_radec, obs_xyz, scans = LSTElevationPlot.load_ms_fields(msname, verbose=1)
        if options.list:
            continue

        skyplot.make_figure(field_time, field_radec, obs_xyz, scans,
                            suptitle=msname, save=options.output_name, display=options.display)

    if options.display and not options.list:
        from pylab import plt
        plt.show()



# -*- coding: utf-8 -*-

from math import *


def radec_to_lmn(ra, dec, ra0, dec0):
    """Returns l,m,n corresponding to direction ra,dec w.r.t. direction ra0,dec0"""
    ## our old formula, perhaps unjustly suspected by me
    ## See purrlog for 3C147_spw0, entries of Nov 21.
    ## Doesn't this break down at the pole (l always 0)?
    l = cos(dec) * sin(ra - ra0)
    m = sin(dec) * cos(dec0) - cos(dec) * sin(dec0) * cos(ra - ra0)
    ## Sarod's formula from LSM.common_utils. Doesn't seem to work right!
    ## (that's because it's for NCP lm coordinates used in NEWSTAR sky models)
    #  l = sin(ra-ra0)*math.cos(dec)
    #   sind0 = sin(dec0)
    #   if sind0 != 0:
    #     m = -(cos(ra-ra0)*cos(dec)-cos(dec0))/math.sin(dec0)
    #   else:
    #     m = 0
    n = sqrt(1 - l * l - m * m)
    return l, m, n


def lm_to_radec(l, m, ra0, dec0):
    """Returns ra,dec corresponding to l,m w.r.t. direction ra0,dec0"""
    # see formula at http://en.wikipedia.org/wiki/Orthographic_projection_(cartography)
    rho = sqrt(l ** 2 + m ** 2)
    cc = asin(rho)

    ra = ra0 + atan2(l * sin(cc), rho * cos(dec0) * cos(cc) - m * sin(dec0) * sin(cc))
    dec = asin(cos(cc) * sin(dec0) + m * sin(cc) * cos(dec0) / rho)

    return ra, dec



#!/usr/bin/env python

import os
from setuptools import setup

install_requires = [

      'astropy',
      'numpy',
      'matplotlib',
      'python-casacore',
      'meqtrees_cattery',
      'scipy',
      'astro-kittens',
      'future-fstrings',
      'six',
      'future',
      'bokeh'
]

setup(name='owlcat',
      version='1.7.6',
      python_requires='>=3.6.0',
      description='miscellaneous utility scripts for manipulating radio interferometry data',
      author='Oleg Smirnov',
      author_email='Oleg Smirnov <osmirnov@gmail.com>',
      url='https://github.com/ska-sa/owlcat',
      packages=['Owlcat'],
      install_requires=install_requires,
      scripts=['Owlcat/bin/' + i for i in os.listdir('Owlcat/bin')],
      data_files=[('Owlcat/bin/', ['Owlcat/bin/commands.list'])],
      )



# -*- coding: utf-8 -*-
#
# % $Id: __init__.py 6998 2009-06-26 08:51:15Z cwilliams $
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import __main__
setattr(__main__, "_meow_verbosity", 0)

from astropy.io import fits as pyfits

from casacore.tables import table, tablecopy, tableexists, tabledelete, addImagingColumns

# pyrap likes to display a lot of these, so shut them off
import warnings

warnings.simplefilter('ignore', DeprecationWarning)

# list of optional packages which will be added to the include path
_Packages = ["Cattery"]

# list of locations where packages will be searched for
_PackageLocations = ["~", "~/Frameworks/",
                     "/usr/local/MeqTrees", "/usr/local/lib/MeqTrees", "/usr/lib/MeqTrees", "/usr/lib64/MeqTrees",
                     "/usr/lib32/MeqTrees"
                     "/usr/local/meqtrees", "/usr/local/lib/meqtrees", "/usr/lib/meqtrees", "/usr/lib64/meqtrees",
                     "/usr/lib32/meqtrees"
                     ]

# mapping of package: path. Filled in as we find packages
_packages = {}

import sys
import os
import os.path


def packages():
    """Returns mapping of available packages to their paths"""
    return _packages
    # print "Using %s, set the %s_PATH environment variable to override this."%(path,package.upper())


def _tryPackageDir(path, package):
    """Tests if path refers to a valid directory, adds it to system include path if so.
    Marks package as having this path."""
    if os.path.isdir(path):
        sys.path.insert(0, path)
        # check for version info
        try:
            version = ' '.join(open(os.path.join(path, 'version_info')))
        except:
            version = 'no version info'
        # insert into packages
        global _packages
        _packages[package] = path, version
        return True
    return False


def _setPackagePath(package):
    """Finds the given package, by first looking in $MEQTREES_PACKAGE_PATH, then checking for
    subdirectories of the standard _PackageLocations list."""
    # check for explicit MEQTREES_PACKAGE_PATH first
    varname = 'MEQTREES_%s_PATH' % package.upper()
    path = os.environ.get(varname, None)
    if path:
        if not _tryPackageDir(path, package):
            print(("Warning: your %s environment variable is set to" % varname))
            print(("%s, but this is not a valid directory." % path))
            print(("The %s package will not be available." % package))
        return
    # else look in standard places
    for path in _PackageLocations:
        path = os.path.expanduser(path)
        if _tryPackageDir(os.path.join(path, package), package):
            return
    # none found
    print(("Warning: No %s package found." % package))
    print(("If you have %s in a non-standard location, please set the %s environment" % (package, varname)))
    print("variable to point to it.")


for pkg in _Packages:
    _setPackagePath(pkg)


def find_exec(execname):
    import os
    path = os.environ.get('PATH') or os.defpath
    for dirname in path.split(os.pathsep):
        fname = os.path.join(dirname, execname)
        if os.access(fname, os.R_OK | os.X_OK):
            return fname
    return None



#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time
import math
import numpy
import Owlcat
from functools import reduce

ROWCHUNK = 50000

time_start = time.time()

from casacore.tables import table, tablecopy, tableexists, tabledelete


def timestamp(format="%H:%M:%S:"):
    return time.strftime(format, time.gmtime(time.time() - time_start))


def progress(message, newline=True):
    sys.stdout.write("%s %-60s%s" % (timestamp(), message, '\n' if newline else '\r'))
    sys.stdout.flush()


if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser

    parser = OptionParser(usage="""%prog: [options] MS""",
                          description="""Finds redundant baselines and sets the WEIGHT (or
IMAGING_WEIGHT) column of the MS, weighting the redundant baselines as 1/n. For a redundant array such as WSRT, this considerably reduces sidelobes in the image.""")
    # parser.add_option("-o","--output",dest="output",type="string",
    # help="name of output FITS file")
    # parser.add_option("-z","--zoom",dest="output_quad",type="string",
    # help="name of zoomed output FITS file")
    parser.add_option("-t", "--tolerance", dest="tolerance", type="float",
                      help="How close (in meters) two baselines need to be to each other to be considered redundant (default .1)")
    parser.add_option("-w", "--weight", action="store_true",
                      help="If this option is given, the IMAGING_WEIGHT column will be "
                           "scaled by 1/n (where n is the number of redundant baselines). The "
                           "default behaviour is to simply set the WEIGHT column to 1/n.")
    parser.add_option("-I", "--ifrs", dest="ifrs", type="string",
                      help="subset of interferometers to use.")
    parser.add_option("-s", "--select", dest="select", action="store",
                      help="additional TaQL selection string. Note that redundant baselines are counted only within the subset "
                           "given by the --ifrs and --select options.")
    parser.add_option("-l", "--list", action="store_true",
                      help="list all baselines and exit.")
    parser.add_option("-r", "--reset", action="store_true",
                      help="reset WEIGHT column to unity. All other options are then ignored.")
    parser.set_defaults(tolerance=.1, select="", ifrs="")

    (options, msnames) = parser.parse_args()

    if len(msnames) != 1:
        parser.error("MS name not supplied.")

    msname = msnames[0]
    ms = table(msname, readonly=False)
    ncorr = ms.getcol('DATA', 0, 1).shape[2]

    if options.reset:
        # apply weights
        nrows = ms.nrows()
        for row0 in range(0, nrows, ROWCHUNK):
            progress("Resetting weights, row %d of %d" % (row0, nrows), newline=False)
            row1 = min(nrows, row0 + ROWCHUNK)
            w = numpy.ones((row1 - row0, ncorr), float)
            ms.putcol('WEIGHT', w, row0, row1 - row0)
        progress("Weights reset to unity.")
    else:
        taqls = []
        # get IFR set
        from Cattery.Meow import IfrSet

        ifrset = IfrSet.from_ms(ms)
        if options.ifrs:
            ifrset = ifrset.subset(options.ifrs)
            taqls.append(ifrset.taql_string())

        if options.select:
            taqls.append(options.select)

        if taqls:
            select = "( " + " ) &&( ".join(taqls) + " )"
            progress("Applying TaQL selection %s" % select, newline=True)
            ms = ms.query(select)
        progress("Looking for redundant baselines", newline=True)
        ant1 = ms.getcol('ANTENNA1')
        ant2 = ms.getcol('ANTENNA2')

        IFRS = sorted(set([(p, q) for p, q in zip(ant1, ant2)]))
        print(("%d baselines" % len(IFRS)))
        groups = []

        for i, (p, q) in enumerate(IFRS):
            bl = ifrset.baseline_vector(p, q)
            # see if this baseline is within the tolerance of a previous group's baseline
            for ig, (bl0, members) in enumerate(groups):
                if abs(bl - bl0).max() < options.tolerance:
                    members.append((p, q))
                    break
            # if none, start a new group
            else:
                members = [(p, q)]
                ig = len(groups)
                groups.append([])
            # update baseline (as mean baseline of group)
            length = reduce(lambda x, y: x + y, [ifrset.baseline_vector(*ifr) for ifr in members]) / len(members)
            groups[ig] = (length, members)

        # convert to list of length,members, and sort by length
        groups = [(math.sqrt((baseline ** 2).sum()), members) for baseline, members in groups]
        groups.sort()

        if options.list:
            baselist = ["%dm (%s)" % (round(length), " ".join(["%s-%s" % (p, q) for p, q in mem]))
                        for length, mem in groups]
            print(("Found %d non-redundant baselines:" % len(baselist), ", ".join(baselist)))
            sys.exit(0)

        # make a dictionary of per-IFR weights
        have_redundancy = False
        weight = dict([((p, q), 1.) for p, q in IFRS])
        for baseline, members in groups:
            if len(members) > 1:
                print(("Baseline %dm, %d ifrs: %s" % (round(baseline), len(members),
                                                     " ".join(["%d-%d" % (p, q) for p, q in members]))))
                have_redundancy = True
                for p, q in members:
                    weight[p, q] = 1.0 / len(members)

        if not have_redundancy:
            print("No redundant baselines found, nothing to do.")
            sys.exit(0)

        # apply weights
        nrows = ms.nrows()
        for row0 in range(0, nrows, ROWCHUNK):
            progress("Applying weights, row %d of %d" % (row0, nrows), newline=False)
            row1 = min(nrows, row0 + ROWCHUNK)
            if not options.weight:
                w = numpy.zeros((row1 - row0, ncorr), float)
                for i in range(row0, row1):
                    w[i - row0, :] = weight[ant1[i], ant2[i]]
                ms.putcol('WEIGHT', w, row0, row1 - row0)
            else:
                w = ms.getcol('IMAGING_WEIGHT', row0, row1 - row0)
                for i in range(row0, row1):
                    w[i - row0, :] *= weight[ant1[i], ant2[i]]
                ms.putcol('IMAGING_WEIGHT', w, row0, row1 - row0)

    progress("Closing MS.")
    ms.close()
